from __future__ import print_function
#  Name   :treedict
#
#          base class for all model_tree_object
#
from __future__ import print_function
#
#
#  Author :
#         Syun'ichi Shiraiwa
#  E-mail :
#         shiraiwa@psfc.mit.edu
#
#
#  History :
#         was born sometime around 2012. 04
#         2012. 08 still keeps growing...
#
# *******************************************
#     Copyright(c) 2012- S. Shiraiwa
# *******************************************

import collections
import weakref
import threading
import traceback
import ifigure.utils.cbook as cbook
import sys
import os
import shutil
import tarfile
import time
import wx
import six
import ifigure.utils.pickle_wrapper as pickle
import ifigure
from ifigure.utils.debug import dprint
import ifigure.ifigure_config as ifigure_config
import ifigure.events
import ifigure.widgets.dialog as dialog
import numpy as np
from ifigure.utils.wx3to4 import deref_proxy

import ifigure.utils.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('TreeDict')

td_name_space = {}


def fill_td_name_space(filename):
    try:
        file, co, mtime = cbook.LoadScriptFile(filename)
        exec(co, {}, td_name_space)
    except:
        print('Failed to build script name space')
        print(traceback.format_exc())


def working_dir_exists_message1(path):
    m = 'Working directory already exists. Do you want to delete it to continue?\nwdir = ' + \
        str(path)
    return m


def working_dir_exists_message2(path):
    m = 'Working directory already exists. Do you want to delete it to continue?\nOtherwise, it will use the existing working directory.\nwdir = ' + \
        str(path) + \
        '\n[note: this could happen when project was not closed properly.]'
    return m


class str_td(str):
    def __init__(self, *args, **kargs):
        str.__init__(self)
        self._td = ''

    @property
    def td(self):
        return self._td

    @td.setter
    def td(self, value):
        self._td = value

# this metaclass forces to read class image file for each subclass


class BaseAttr(type):
    def __init__(cls, name, base, d):
        cls._image_load_done = False
        type.__init__(cls, name, base, d)


class TreeDict(object):
    # index number of object commonly used
    # among all classes derived from TreeDict
    __metaclass__ = BaseAttr
    _debug = 0
    _id_base = 0
    _save_var = True
    # _image_load_done=False
    _image_id = [-1]

    def __new__(cls, *args, **kargs):
        obj = object.__new__(cls)
        if not hasattr(cls, '_image_load_done'):
            cls._image_load_done = False
        # subclass's __new__ may be overwritten to implement
        # an extra procees such as the cancelation of object
        # creation. In such case, obj members may be edited
        # in subclass's __new__. Variables which is editable
        # during __new__ should be initialized here.

        # self._var stores generic variable "visible from user"
        # attribute variable
        obj._var = collections.OrderedDict()
        # description for attribute
        obj._note = collections.OrderedDict()

        # nodo variable is mostly
        # for subtree contents
        # if node vairialbe is used, _can_have_child must be
        # false
        obj._var0 = None
        obj._var0_show = False
        obj._name = ''
        obj._parent = None
        obj._nosave_var0 = False
        obj._can_have_child = True
        obj._has_private_owndir = True
        obj._genuine_name = ''     # this is the name shown in tree viewer
        # it is called genuine since when external folder
        # is loaded, this is the name of the file.
        # OTH, _name might be different to keep the project
        # tree structure valid
        # In such case, the object can not be renamed
        # self._d stores children
        obj._d = []
        return obj

    def __init__(self, parent=None, src=None):
        self._id = TreeDict._id_base
        self._name = 'treedict'

        if parent is not None:
            dprint2('treedict::__init__ is called with parent', parent)
        self.set_parent(parent)
        TreeDict._id_base = TreeDict._id_base+1
        self._status = ''
        self._items = []
        self._suppress = False
        # store if it is visible in tree viewer
        self._visible = False
        self._use_custom_image = False
        self._custom_image_load_done = False
        self._image_update = False
        self._custom_image_id = [-1]
        self._name_readonly = False
        self.__initialised = True
        self._save_mode = 0
        self._var_changed = False  # a flag to judge the need of updating var viewer

    def __repr__(self):
        return self.__class__.__name__+'('+self.get_full_path()+')'

    def compact_name(self):
        txt = self.get_full_path()
        if len(txt) < 32:
            return txt
        else:
            names = [x for x in reversed(
                [t._name for t in self.walk_tree_up()])]
            for x in range(len(names)-2):
                names[x+1] = '.'
                if len('.'.join(names)) < 32:
                    return '.'.join(names)
        return self.get_root_parent().name + '....' + self.name

    @classmethod
    def isTreeDict(self):
        return True

    @classmethod
    def isTreeLink(self):
        return False

    @classmethod
    def get_namebase(self):
        return 'treedict'

    @classmethod
    def pv_kshortcut(self, key):
        '''
        define keyborad short cut on project viewer
        must return a method responding key
        '''
        return None

    def can_have_child(self, child=None):
        return self._can_have_child

    def get_can_have_child(self):
        return self._can_have_child

    def set_can_have_child(self, value):
        self._can_have_child = value

    def ref(self):
        # this is to get self from weakref proxy object
        # additionally, proxy alive check uses a call
        # to this method....
        return self

    def classimage(self):
        if self._use_custom_image:
            return self.customimage()
        else:
            return self.classimage0()

    def get_classimage(self):
        img = self.classimage()
        if not self._suppress:
            return img[0]
        else:
            return img[1]

    @classmethod
    def classimage0(cls):
        if cls._image_load_done is False:
            cls._image_id = cls.load_classimage()
            cls._image_load_done = True
        return cls._image_id[0]

    def customimage(self):
        if self._custom_image_load_done is False:
            self._custom_image_id = self.load_customimage()
            self._custom_image_load_done = True
        if self._custom_image_id[0] == -1:
            return self.classimage0()
        return self._custom_image_id[0]

    @classmethod
    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx = cbook.LoadImageFile(path, 'no_icon.png')
        return [idx]

    @classmethod
    def is_movable(self):
        #  [0]: allow to change order under the same parent
        #  [1]: allow to move other place
        return (False, False)

    def destroy(self, clean_owndir=True):
        for name, child in self.get_children():
            child.destroy(clean_owndir=clean_owndir)
        # clean workdir
        if clean_owndir:
            if self._has_private_owndir:
                self.rm_owndir()
            else:
                self.clean_owndir()

        if self._parent is not None:
            self._parent._del_child(self)

        # clean var
        for key in self._var:
            self._var[key] = None

    def __getattr__(self, attr):
      # only attributes not starting with "_" are organinzed
      # in the tree
        if not attr.startswith("_"):
            idx = self.i_child(name=attr)
            if idx != -1:
                return self._d[idx]
            # print "making attr", attr
            # print self._name, self.get_full_path(), attr
            raise AttributeError('Tree member not found :'+attr+':'+str(self))
            # return self._d.setdefault(attr, TreeDict())
        raise AttributeError('Tree name should not start from "_"')

    def __setattr__(self, attr, val):
        if '_TreeDict__initialised' not in self.__dict__:
           #         this test allows attributes to be set in the __init__ method
            return object.__setattr__(self, attr, val)
        elif attr.startswith("_"):
            return object.__setattr__(self, attr, val)
        try:
            if val.isTreeDict():
                self.add_child(attr, val)
        except Exception:
            raise AttributeError("Do not make attribut "+attr)

    def __getstate__(self):
            # for pickling
        return self._d, self._attributes()

    def __getitem__(self, key):
        if self._var0 is not None:
            try:
                return self._var0[key]
            except:
                return self._var0
#           if isinstance(self._var0, collections.Hashable):
#               return self._var0[key]
#           else:
#               return self._var0
        else:
            raise TypeError('Object does not have var0')

    def __setitem__(self, key, value):
        # slicing weakref with [:] gives the 2nd...why?
        if (key == slice(None, None, None) or
                key == slice(0, 2147483647, None)):
            self._var0 = value
            ifigure.events.SendVar0ChangedEvent(self)
        else:
            if self._var0 is None:
                raise KeyError('Object does not have ver0')
            self._var0[key] = value

    def __setstate__(self, tp):
        # for unpickling
        d, l = tp
        self._d = d
        for name, obj in l:
            setattr(self, name, obj)

    def _process_kywd(self, kywds, key, defv):
        #        print 'in', kywds
        if key in kywds:
            self.setvar(key, kywds[key])
            del kywds[key]
        else:
            self.setvar(key, defv)
#        print 'out', kywds
        return kywds

    def _process_kywd2(self, kywds, key, defv):
        #        print 'in', kywds
        if key in kywds:
            val = kywds[key]
            del kywds[key]
        else:
            val = defv
#        print 'out', kywds
        return kywds, val

    def list_all(self):
        # easy to read string representation of data
        # print "generating string"
        rl = []
        
        from six import iteritems        
        for k, v in iteritems(self._getLeaves()):
            rl.append("%s = %s" % (k, v.__repr__()))
   #        return "\n".join(rl)
        return rl

    def _getLeaves(self, prefix=""):
        # getLeaves tree, starting with self
        # prefix stores name of tree node above
        prefix = prefix and prefix + "."
        rv = {}
        atl = self.get_childnames()
        for at in atl:
            ob = self.get_child(name=at)
            trv = ob._getLeaves(prefix+at)
            rv.update(trv)

        for at, ob in self._attributes():
            rv[prefix+at] = ob
        return rv

    def _attributes(self):
        # return 'lenaves' of the data tree
        # return self._d.items()
        return [(s._name, s) for s in self._d]
    #
    #  access routine to children
    #

    def get_childnames(self):
        names = []
        for child in self._d:
            names.append(child._name)
        return names

    def has_child(self, name):
        return self.get_childnames().count(name) != 0

    def get_children(self):
        return self._attributes()

    def num_child(self):
        return len(self._d)

    def get_ichild(self):
        i = self._parent.i_child(self)
        return i

    def i_child(self, obj=None, name=None):
        if name is not None:
            keys = self.get_childnames()
            if keys.count(name) == 0:
                return -1
            return keys.index(name)
        if obj is not None:
            #c = self._d.count(obj)
            #if c == 0:
            #    return -1
            try:
                idx = self._d.index(obj)
                return idx
            except ValueError:
                return -1
            #return idx
        return -1
    
    def get_child(self, idx=None, name=None):
        if idx is not None:
            if len(self._d) > idx > -1:
                return self._d[idx]
            else:
                return None
        if name is not None:
            names = self.get_childnames()
            c = names.count(name)
            if c == 0:
                return None
            idx = names.index(name)
            return self._d[idx]
        return None

    def get_first_child(self):
        return self.get_child(0)

    def get_last_child(self):
        return self.get_child(self.num_child()-1)

    def get_next_child(self, child):
        try:
            return self._d[self._d.index(child)]
        except:
            return None
#        idx = self.i_child(child)
#        if idx == -1 : return None
#        return self.get_child(idx+1)

    def get_prev_child(self, child):
        idx = self.i_child(child)
        if idx == -1:
            return None
        if idx == 0:
            return None
        return self.get_child(idx-1)

    def get_depth(self):
        n = 0
        p = self
        while p is not None:
            p = p.get_parent()
            n = n+1
        return n

    def isdescendant(self, kid):
        p = kid
        while p is not None:
            if self == p:
                return True
            p = p.get_parent()
        return False

    def get_relative_path(self, td):
        dprint1('get_relative_path is going to be obsolete, consider get_td_path')
        # find common parent
        k = 0
        p = self
        while p is not None:
            if p.isdescendant(td):
                break
            p = p.get_parent()
            k = k + 1
#        print 'common parent', p.get_full_path()
        a = [i for i in reversed([t._name for t in td.walk_tree_up()])]
        b = [i for i in reversed([t._name for t in p.walk_tree_up()])]
        return [".."]*k + a[len(b):]

    def get_td_path(self, td, fullpath=False):
        if fullpath:
            return td.get_full_path().split('.')
        # find common parent
        k = 0
        p = self
        while p is not None:
            if p.isdescendant(td):
                break
            p = p.get_parent()
            k = k + 1
#        print 'common parent', p.get_full_path()
        a = [i for i in reversed([t._name for t in td.walk_tree_up()])]
        b = [i for i in reversed([t._name for t in p.walk_tree_up()])]
        return ['.']+[".."]*k + a[len(b):]

    def resolve_td_path(self, path):
        # path = [".", "..","..", someting]
        # path = ["..","..", someting]
        if path[0] == '.':
            p = self
        else:
            p = self.get_root_parent()
        p2 = path[1:]

        for name in p2:
            if name == '..':
                p = p.get_parent()
            else:
                p = p.get_child(name=name)
        return p

    def _del_child(self, obj=None, idx=None, name=None):
        #
        #  this is not a fuction to be called directly.
        #  use destroy instead. it does not delete object
        #  instead unregister it from children list
        #
        if obj is not None:
            idx = self.i_child(obj=obj)
            if idx == -1:
                return
        elif name is not None:
            idx = self.i_child(name=name)
            if idx == -1:
                return

        self._d[idx].set_parent(None)
        del self._d[idx]

    def add_child(self, *args, **kargs):
        '''
           add_child(name, obj)
           add_child(obj) : name is not changed
           check potential conflict of method and child name
        '''
        from ifigure.utils.cbook import is_safename

        def name_check(newname, warning):
            fpath = self.get_full_path()+'.'+newname
            do_rename = False
            newname = '_'.join(newname.split(' '))
            if '.' in newname:
                newname = newname.split('.')[0]
                if not is_safename(newname):
                    warning.append(fpath + 'is skipped.')
                    return False, newname, do_rename
            if not is_safename(newname):
                if is_safename('renamed_'+newname):
                    newname = 'renamed_'+newname
                    do_rename = True
                    warning.append(fpath + 'is renamed in Tree.')
                else:
                    warning.append(fpath + 'is skipped.')
                    return False, newname, do_rename
            if newname.startswith('_'):
                newname = 'renamed_'+newname
                do_rename = True
                warning.append(fpath + 'is renamed in Tree.')
            return True, newname, do_rename

        print_warn = True if 'warning' in kargs else False
        warning = kargs.pop('warning', [])

        if len(args) == 2:
            name = str(args[0])  # avoid unicode name...
            obj = args[1]
        elif len(args) == 1:
            obj = args[0]
            name = obj.name
        if obj is not None:
            bname = name
            i = 0
            conflict = False
            if bname == '__pycache__':
                #print('skipping ' + bname)
                return
            flag, bname1, do_rename = name_check(bname, warning)
            if not flag:
                print('can not add '+name)
                return
            while hasattr(self, bname1):
                conflict = True
                warning.append(self.get_full_path() +
                               ' already has attr ' + bname1)
                bname1 = bname1 + str(i)
                i = i+1
            obj._genuine_name = bname
            obj.set_parent(self)     # this could raise error so do first
            self._d.append(obj)
            if self.is_suppress():
                obj.set_suppress(self.is_suppress())
            obj._name = bname1

            if conflict:
                warning[-1] = warning[-1]+'. Renamed to ' + bname

            #if print_warn: print('\n'.join(warning))
            return len(self._d)-1  # == self.i_child(obj)

        print("Empty object can not be added")

    def move(self, new_parent, keep_zorder=False):
        '''
        def move(self, new_parent, keep_zorder=False)
        move object under a new parent'
        '''
        p = self.get_parent()
        odir = self.owndir()
        if p is None:
            return
        p._del_child(obj=self)

        name = self.name
        if new_parent.get_childnames().count(name) != 0:
            name = new_parent.get_next_name(name)
        new_parent.add_child(name, self, keep_zorder=keep_zorder)

        ndir = self.owndir()
        if self._has_private_owndir:
            #        if self._can_have_child:
            #            dprint1('moving ', odir, ndir)
            if os.path.exists(odir):
                if not self.has_owndir():
                    self.mk_owndir()
                os.rename(odir, ndir)
#                shutil.move(odir, ndir)
        else:
            for item in set(self.ownitem()):
                shutil.move(os.path.join(odir, item),
                            os.path.join(ndir, item))

    def move_child(self, idx, idx2):
        if idx == idx2:
            # do nothing
            return
        item = self._d
        item = cbook.MoveItemInList(item, idx, idx2)
        self._d = item

    def find_by_id(self, id):
        root = self.get_root_parent()
        for obj in root.walk_tree():
            if obj.id == id:
                return obj
        return None

    def find_by_full_path(self, name):
        a = name.split('.')
        root = self.get_root_parent()
        if a[0] != root.name:
            return None
        p = root
        for name in a[1:]:
            p = p.get_child(name=name)
            if p is None:
                print(('can not find', name))
                return None
        return p

    def expand_path(self, name):
        '''
           expand path string like 'proj.model.*.scripts'
           and returns list of treedict
        '''
        a = name.split('.')
        root = self.get_root_parent()
        if a[0] != root.name:
            return []
        arr = [root]
        for name in a[1:]:
            arr2 = []
            for k, p in enumerate(arr):
                if name == '*':
                    arr2.extend(list(zip(*p.get_children())[1]))
                else:
                    p = p.get_child(name=name)
                    if p is not None:
                        arr2.append(p)
            arr = arr2
        return arr

    def set_parent(self, parent):
        self._parent = parent

    def get_parent(self):
        return self._parent

    def get_app(self):
        root = self.get_root_parent()
        app = root.app if hasattr(root, 'app') else None
        app = deref_proxy(app)
        return app

    def get_root_parent(self):
        if self.get_parent() is None:
            return self
        else:
            return (self.get_parent()).get_root_parent()

    def get_grand_parent(self):
        # return the parent of parent
        p = self.get_parent()
        if p is None:
            return None
        return p.get_parent()

    def get_full_path(self):
        names = reversed([t._name for t in self.walk_tree_up()])
        return '.'.join(names)
#        return self._do_get_full_path('')

    def _do_get_full_path(self, base):
        if base == '':
            base = self._name
        else:
            base = self._name+'.'+base
        if self._parent is not None:
            base = self._parent._do_get_full_path(base)
        return base

    def walk_tree(self, stop_at_ext=False):
        '''
        generator to wolk through tree
        '''
        from ifigure.mto.py_extfile import ExtMixIn
        yield self
        if stop_at_ext and isinstance(self, ExtMixIn):
            return
#       child=self.get_first_child()
        if len(self._d) == 0:
            return
        child = self._d[0]
        k = 0
        while child is not None:
            for x in child.walk_tree(stop_at_ext=stop_at_ext):
                yield x
#           child=self.get_next_child(child)
            k = k+1
            if len(self._d) == k:
                break
            child = self._d[k]

    def walk_tree_rev(self):
        '''
        generator to wolk through tree (reversed)
        '''
        child = self.get_last_child()
        while child is not None:
            for x in child.walk_tree_rev():
                yield x
            child = self.get_prev_child(child)
        yield self

    def walk_tree_up(self, cond=True):
        ''' 
        generator to walk tree upward from
        self. if cond is given, generator
        stops when the condition is satisefied.
        cond should  takes arguments, to which
        self is 
        '''
        k = cond
        if callable(cond):
            k = cond(self)
        if k:
            yield self
            if self.get_parent() is not None:
                for x in self.get_parent().walk_tree_up(cond=cond):
                    yield x

    def walk_tree_down(self, cond=True, top=None):
        ''' 
        generator to walk tree downward from
        top to self. if cond is given, generator
        stops when the condition is satisefied
        '''
        tmp = []
        for obj in self.walk_tree_up(cond=cond):
            tmp.append(obj)
            if tmp is top:
                break

        tmp2 = reversed(tmp)
        for item in tmp2:
            yield item

    def is_descendant(self, kid):
      # ture if kid is self's offspring
        p = kid.get_parent()
        while p is not None:
            if self == p:
                return True
            p = p.get_parent()
        return False

    def is_ascendant(self, parent):
        return parent.is_descendant(self)

    def find_common_parent(self, td_list):
        p = self
        while p is not None:
            flag = True
            for t in td_list:
                if t is None:
                    return self.get_root_parent()
                if p is t:
                    break
                if not p.is_descendant(t):
                    flag = False
                    break
            if flag:
                return p
            p = p.get_parent()
        return p
   #
   #   Variable/Note Setter and Getter
   #

    def get_varlist(self):
        return list(self._var)

    def hasvar(self, name):
        return name in self._var

    def setvar(self, *args):
        #       print self, args
        self._var_changed = True
        if len(args) == 2:
            self._var[args[0]] = args[1]
        if len(args) == 1:
            if isinstance(args[0], dict):
                self._var = args[0]

    def dynamic_names(self):
        return [key for key in self._var
                if cbook.isdynamic(self._var[key])]

    def eval(self, *args, **kargs):
        '''
        eval values in _var
        eval("x")  = evaluate "x"
        eval("x", "y", "z") = evaluate "x", "y", "z"
        eval() = return self._var0
        '''
        if 'np' in kargs:
            np = kargs['np']
        else:
            np = False
        if len(args) == 1:
            return self._eval(args[0], use_np=np)
        if len(args) == 0:
            return self._eval()
        return (self._eval(name, use_np=np) for name in args)

    def get_pyfolder(self):
        from ifigure.mto.py_code import PyFolder
        p = self
        while p is not None:
            if isinstance(p, PyFolder):
                return p
            p = p.get_parent()
        return None

    def get_pymodel(self):
        from ifigure.mto.py_code import PyModel
        p = self
        while p is not None:
            if isinstance(p, PyModel):
                return p
            p = p.get_parent()
        return None

    def get_figbook(self):
        from ifigure.mto.fig_book import FigBook
        p = self
        while p is not None:
            if isinstance(p, FigBook):
                return p
            p = p.get_parent()
        return None

    def get_figpage(self):
        from ifigure.mto.fig_page import FigPage
        p = self
        while p is not None:
            if isinstance(p, FigPage):
                return p
            p = p.get_parent()
        return None

    def get_figaxes(self):
        from ifigure.mto.fig_axes import FigAxes
        p = self
        while p is not None:
            if isinstance(p, FigAxes):
                return p
            p = p.get_parent()
        return None

    def get_variables(self):
        from ifigure.mto.py_code import AbsNamespacePointer
        p = self
        while p is not None:
            if isinstance(p, AbsNamespacePointer):
                return p
            p = p.get_parent()
        return None

    def get_namespace(self):
        p = self.get_variables()
        if p is None:
            return {}
        if p.has_ns():
            return p.get_ns()._var
        else:
            return {}

    def get_extfolderpath(self):
        from ifigure.mto.py_code import AbsFileContainer
        p = self
        namelist = []
        while p is not None:
            if isinstance(p, AbsFileContainer):
                if p.getvar('ext_folder') is not None:
                    namelist.append(p.getvar('ext_folder'))
                    return os.path.join(*list(reversed(namelist)))
                namelist.append(p._genuine_name)
            p = p.get_parent()

    def get_extfolderpathfolder(self):
        from ifigure.mto.py_code import AbsFileContainer
        p = self
        namelist = []
        while p is not None:
            if isinstance(p, AbsFileContainer):
                if p.getvar('ext_folder') is not None:
                    return p
            p = p.get_parent()

    def _eval(self, name='', use_np=False):
        try:
            root = self.get_root_parent()
            lc = {key: td_name_space[key] for key in td_name_space}
            lc[root.name] = root
            m = self.get_pymodel()
            if m is not None:
                lc['model'] = m
            if m is not None:
                lc['param'] = m.param
            m = self.get_figpage()
            if m is not None:
                lc['page'] = m
            m = self.get_figbook()
            if m is not None:
                lc['book'] = m

            if six.PY2 and isinstance(self._var[name], unicode):
                if self._var[name].startswith(u'='):
                    self._var[name] = str(self._var[name])
            if isinstance(self._var[name], str):
                if name == '':
                    value = self._var0
                    return self._var0
                if cbook.isdynamic(self._var[name]):
                    value = eval((self._var[name])[1:], lc, lc)
#                  value = eval()
                    if use_np and not isinstance(value, np.ndarray):
                        if value is None:
                            return None
                        if hasattr(value, '__iter__'):
                            value = np.array(value)
                            if value.ndim == 0:
                                return None
                            return value
                        else:
                            return value
                    return value
                return self._var[name]
            value = self._var[name]
            if use_np and not isinstance(value, np.ndarray):
                if value is None:
                    return None
                if hasattr(value, '__iter__'):
                    value = np.array(value)
                    if value.ndim == 0:
                        return None
                    return value
                else:
                    return value
            return value
        except Exception:
            print(sys.exc_info())
#          pass
        return None

    def getvar(self, *args):
        if len(args) == 0:
            return self._var
        try:
            if len(args) == 1:
                return self._var[args[0]]
            else:
                return tuple([(self._var[name] if name in self._var else None) for name in args])
        except KeyError:
            if len(args) == 1:
                return None
            return [None]*len(args)

    def getvar_copy(self):
        var = self.getvar()
        d = collections.OrderedDict()
        for key in var:
            d[key] = var[key]
        return d

    def compare_vars(self, td):
        """
        compare vars stored in two tree dicts
        """
        print(("comparing", self.get_full_path(), td.get_full_path()))
        diff = cbook.DictDiffer(self.getvar(), td.getvar())

    def setnote(self, *args):
        '''
        setnote('xxx', 'xxx is yyy')
        '''
        if len(args) == 2:
            try:
                a = self._var[args[0]]
            except KeyError:
                print(("!!!", self.get_full_path(), " does not have ", name))
                return
            self._note[args[0]] = args[1]
        if len(args) == 1:
            self._note = args[0]

    def getnote(self, name=None):
        if name is None:
            return self._note
        try:
            return self._note[name]
        except KeyError:
            return ''

    def delvar(self, name):
        try:
            del self._var[name]
        except KeyError:
            pass
        try:
            del self._note[name]
        except KeyError:
            pass

    #
    #   methods relating to var0
    #
    def setvar0(self, value):
        self._var0 = value
        ifigure.events.SendVar0ChangedEvent(self)

    def getvar0(self): return self[:]

    def get_contents(self, *args, **kargs):
        from ifigure.mto.py_contents import PyContents
        if not isinstance(self._var0, PyContents):
            return None
        contents = self._var0
        for key in args:
            contents = contents[key]
        if hasattr(contents, 'eval'):
            return contents.eval(self, **kargs)
        return contents

    def set_contents(self, *args):
        from ifigure.mto.py_contents import PyContents
        if not isinstance(self._var0, PyContents):
            return None
        contents = self._var0
        if len(args) > 2:
            for key in args[:-2]:
                contents = contents[key]
        contents[args[-2]] = args[-1]

    def get_drag_text2(self, key):
        ''' 
        text for dnd from var viewer
        '''
        return self.get_full_path()+'.eval("'+key+'")'

    #
    #  suppress
    #
    def set_suppress(self, val):
        p = self.get_parent()
        if p is None:
            return
        if (p.is_suppress() and val == False):
            print('parent object is suppressed, can not unsuppress the object')
            return
        self._suppress = val
        self._image_update = False
        for name, child in self.get_children():
            child.set_suppress(val)

    def is_suppress(self):
        return self._suppress

    @property
    def isSuppressed(self):
        return self._suppress

    #
    #  visibility in tree viewer
    #
    def set_visible(self, val):
        self._visible = val

    def is_visible(self):
        return self._visible

    #
    #   menu in tree viewer
    #
    def tree_viewer_menu(self):
        # return MenuString, Handler, MenuImage
        if not wx.GetApp().TopWindow.appearanceconfig.setting['show_suppress_menu']:
            menu = []
        else:
            if self._suppress:
                menu = [('-Suppress',   self.onSuppress, None),
                        ('Unsuppress', self.onUnSuppress, None),
                        ('Delete...', self.onDelete, None)]
            else:
                menu = [('Suppress',   self.onSuppress, None),
                        ('-Unsuppress', self.onUnSuppress, None),
                        ('Delete...', self.onDelete, None)]

        if self._name_readonly:
            menu.append(('-Rename...', self.onRename, None))
        else:
            menu.append(('Rename...', self.onRename, None))
        from ifigure.mto.py_extfile import ExtMixIn
        if not isinstance(self, ExtMixIn):
            from ifigure.mto.hg_support import has_hg
            if self.can_have_child():
                menu.append(('+Import Subtree', None, None))
                mm = [('From File...', self.onLoadSubTree, None), ]
                if has_hg:
                    mm.append(('From Repository...', self.onLoadSubTreeHG,
                               None))
                if hasattr(self, 'onAddExtFolder'):
                    mm.append(('From Folder...', self.onAddExtFolder, None))
                menu.extend(mm)
                menu.append(('!', None, None))
                if (self.hasvar('subtree_path') and
                        os.path.exists(self.getvar('subtree_path'))):
                    menu.append(('Export Subtree', self.onSaveSubTree, None))
                else:
                    menu.append(('-Export Subtree', None, None))
                menu.append(('+Export As...', None, None))
                menu.append(('Subtree...', self.onSaveSubTreeAs, None))
                if self.has_owndir():
                    menu.append(('Files...', self.onExportFiles, None))
                menu.append(('!', None, None))
            else:
                if (self.hasvar('subtree_path') and
                        os.path.exists(self.getvar('subtree_path'))):
                    menu.append(
                        ('Export '+self._name, self.onSaveSubTree, None))
                else:
                    menu.append(('-Export '+self._name, None, None))
                menu.append(('Export '+self._name + ' As...',
                             self.onSaveSubTreeAs, None))

        from ifigure.mto.py_code import PyFolder

        if self.can_have_child(PyFolder()):
            if isinstance(self, ExtMixIn):
                menu.append(
                    ('Reload File Tree...', self.onReloadExtTree, None))
                menu.append(
                    ('Change Dir. to here', self.onChangeDirToExt, None))
            menu.append(('New Folder...', self.onAddFolder, None))
        return menu

    def onDelete(self, event):
        self.destroy()
        event.Skip()

    def onSuppress(self, event):
        self.set_suppress(True)
        event.Skip()

    def onUnSuppress(self, event):
        self.set_suppress(False)
        event.Skip()

    def dlg_ask_newname(self):
        app = self.get_app()
        ret, new_name = dialog.textentry(app,
                                         "Enter a new name", "Rename...", self.name)
        if ret:
            p = self.get_parent()
            if p is not None:
                if hasattr(p, new_name):
                    ret = dialog.message(app,
                                         'attribute name (' + new_name +
                                         ') is already used',
                                         'Name Conflict',
                                         0)
                    return None
                if new_name == 'tmp':
                    ret = dialog.message(app,
                                         'attribute name (' +
                                         new_name + ') is reserved',
                                         'Reserved Name',
                                         0)
                    return None
            return new_name

    def onRename(self, event):
        new_name = self.dlg_ask_newname()
        if new_name is not None:
            flag = self.rename(new_name)
            ifigure.events.SendChangedEvent(self)

    def onSaveSubTree(self, e):
        if (self.hasvar('subtree_path') and
                os.path.exists(self.getvar('subtree_path'))):
            path = self.getvar('subtree_path')
            print('exporting subtree to '+path)
            self.save_subtree(path)
        else:
            self.onSaveSubTreeAs(e)

    def onSaveSubTreeAs(self, e):
        path = dialog.write(defaultfile=self.name+'.pfs',
                            wildcard='*.pfs')
        if path == '':
            return
        if path[-4:] != '.pfs':
            path = path+'.pfs'
        print('exporting subtree to '+path)
        self.save_subtree(path)
        self.setvar('subtree_path', path)

    def onExportFiles(self, e):
        path = dialog.writedir(parent=e.GetEventObject(),
                               message='Select directory to save files',)

        if path == '':
            return
        path = os.path.join(path, self._name)
        print('exporting files to '+path)
        import shutil
        shutil.copytree(self.owndir(), path)

    def onLoadSubTree(self, e):
        path = dialog.read(wildcard='*.pfs')
        if path == '':
            return
        child = self.load_subtree(path)
        if child is not None:
            child.setvar('subtree_path', path)
        e.Skip()

    def onLoadSubTreeHG(self, e):
        from ifigure.mto.hg_support import load_subtree_hg
        load_subtree_hg(self)
        e.Skip()

    def onAddFolder(self, event=None, name=''):
        from ifigure.mto.py_code import PyFolder
        app = self.get_app()
        if name == '':
            ret, new_name = dialog.textentry(app,
                                             "Enter Foler Name", "Add Folder...", 'folder')
        else:
            ret = True
            new_name = name
        if ret:
            if self.has_child(new_name):
                ret = dialog.message(app,
                                     new_name + ' is already used',
                                     'Name Conflict',
                                     0)
                return
            obj = PyFolder()
            self.add_child(new_name, obj, keep_zorder=True)
        ifigure.events.SendChangedEvent(self)

    #
    #  utilitiy to add child object
    #
    def add_childobject(self, cls, name, *args, **kargs):
        if self.has_child(name):
            name = self.get_next_name(name)
        try:
            obj = cls(*args, **kargs)
            idx = self.add_child(name, obj, keep_zorder=True)
            o = self.get_child(name=name)
            if o is not None:
                return o
            return self.get_child(idx=idx)
        except:
            import traceback
            traceback.print_exc()
            return

    #
    #   get_new_name
    #
    def get_next_name(self, header):
        keys = self.get_childnames()
        return cbook.GetNextName(keys, header)

    #
    #   rename
    #
    def is_renamable(self):
        return self._name_readonly

    def set_namereadonly(self,  value=True):
        self._name_readonly = value

    def rename(self, new, ignore_name_readonly=False):
        # rename tree object

        # this routine does..
        #   1) check name conflict in tree
        #   2) check owndir conflict

        #
        # root can not be renamed
        from ifigure.utils.cbook import is_safename
        if self._name_readonly and not ignore_name_readonly:
            raise AttributeError("Name read only :" + self.__repr__())
        if not is_safename(new):
            dprint1(new + ' can not be used as variable name')
            return False
        if self._name == new:
            return True
        if self.get_parent() is None:
            return False
        if hasattr(self.get_parent(), new):
            print(self.get_parent().get_full_path() + ' already has attr ' + new)
            return False
        if self._name != self._genuine_name:
            dprint1('Cannot rename since object has different screen name')
            return False

        keys = self.get_parent().get_childnames()
        if keys.count(new) != 0:
            print(' name is already used')
            return False

        rename_owndir = False
        if self.has_owndir():
            od = self.owndir()
            rename_owndir = True
        old = self._name
        self._name = new
        self._genuine_name = new
        ifigure.events.SendSimpleTDEvent(None, td=self, w=None,
                                         useProcessEvent=False,
                                         code='Rename')

        if (self.has_owndir() and  # self._can_have_child
                self._has_private_owndir):
            print(' name seems to be used in the past')
            print(' delete ' + self.has_owndir() + ' manually')
            self._name = old
            return False
        if (rename_owndir and
                self._has_private_owndir):
            #            self._can_have_child):
            os.rename(od, self.owndir())
        return True

    #
    #   own directory....
    #
    def mk_owndir(self):
        if self.has_owndir():
            return
#        if not self._can_have_child:
        if not self._has_private_owndir:
            self._parent.mk_owndir()
            return
        path = self.get_root_parent().getvar("wdir")
        l = self.get_full_path().split('.')
#        if len(l) < 2: return
#        for item in l[1:]:
        for item in l:
            path = os.path.join(path, item)
            if not os.path.exists(path):
                os.mkdir(path)

    def rm_owndir(self):
        #        if not self._can_have_child:return
        if not self._has_private_owndir:
            return
        if self.has_owndir():
            try:
                shutil.rmtree(self.owndir())
            except Exception:
                pass

    def rm(self, file):
        '''
        remove a file in owndir
        '''
        try:
            #            if self.owndir() is None: return
            #            if file is None: return
            if os.path.exists(os.path.join(self.owndir(), file)):
                os.remove(os.path.join(self.owndir(), file))
                ifigure.events.SendFileSystemChangedEvent(self)
        except Exception:
            import traceback
            traceback.print_exc()

    def clean_owndir(self):
        def rmgeneric(path, __func__):
            try:
                __func__(path)
            except OSError:
                print(('Remove error', path))

        def removeall(path):
            if not os.path.isdir(path):
                return
            files = os.listdir(path)
            for x in files:
                fullpath = os.path.join(path, x)
                if os.path.isdir(fullpath):
                    removeall(fullpath)
                    f = os.rmdir
                    rmgeneric(fullpath, f)
                elif os.path.isfile(fullpath):
                    f = os.remove
                    rmgeneric(fullpath, f)
                elif os.path.islink(fullpath):
                    f = os.remove
                    rmgeneric(fullpath, f)

#       if self._can_have_child:
        ifigure.events.SendFileSystemChangedEvent(self)

        if self._has_private_owndir:
            removeall(self.owndir())
        else:
            for item in self.ownitem():
                path = os.path.join(self.owndir(), item)
                if not os.path.exists(path):
                    continue
                if os.path.isdir(path):
                    shutil.rmtree(path)
                else:
                    os.remove(path)

    def owndir(self):
        from ifigure.ifigure_config import rcdir
        wdir = self.get_root_parent().getvar("wdir")
        if self.get_root_parent().getvar("wdir") is None:
            return os.path.join(rcdir, 'Should not exist')
        l = self.get_full_path().split('.')
        if self._has_private_owndir:
            #        if self._can_have_child:
            l2 = [wdir]+l
        else:
            l2 = [wdir]+l[:-1]
        return os.path.join(*l2)

    def has_owndir(self):
        return os.path.exists(self.owndir())

    def ls_owndir(self):
        if not self.has_owndir():
            self.mk_owndir()
        return os.listdir(self.owndir())

    def lsl_owndir(self, command='ls -l'):
        '''
        works only on unix
        '''
        import subprocess as sp
        if not self.has_owndir():
            self.mk_owndir()
        d = os.getcwd()
        print(sp.Popen(command, stdout=sp.PIPE, shell=True).stdout.read())
        os.chdir(d)
    #
    #  write2shell : write variable to shell
    #

    def write2shell(self, value, name):
        '''
        write variable to shell
        '''
        root = self.get_root_parent()
        if not hasattr(root, 'app'):
            return
        if root.app is None:
            return
        flag = True if name in root.app.shell.lvar else False
        root.app.shell.lvar[name] = value
#        txt = 'Data exported as ' + name
#        root.app.shell.push('print %r' % txt, silent = True)
        return flag

    #
    #  save/load
    #     save and load are entry point, actual
    #     save/load action depndent on derived
    #     class is implemented in _do_***_data
    #
    #     note: format version
    #        1: old format, each class can write multiple
    #           pickle data.
    #        2: save_data2/load_data2
    #             save_data2 returns a dictionary to
    #             be pickled once.
    #           each object always does two pickle actions
    #           one for header, the other for data
    #           probably easier to extend.

    def save(self, fid=None, olist=None):
        return self.save2(fid, olist)

    def save_data(self, fid=None):
        h2 = {}
        pickle.dump(h2, fid)

    def save2(self, fid=None, olist=None):
        if olist is None:
            olist = []

        h2 = {"name": self.__class__.__name__,
              "module": self.__class__.__module__,
              "num_child": self.num_child(),
              "sname": self.name,
              "id": self.id,
              "var0": self._var0,
              "note": self._note,
              "format": 2}
        if self._save_var:
            h2["var"] = self._var
#        print h2
        if self._nosave_var0:
            h2["var0"] = None
        if fid is not None:
            pickle.dump(h2, fid)
        c = olist.count(self.id)
        if c == 0:
            data = self.save_data2({})
            pickle.dump(data, fid)
            olist.append(self.id)
        else:
            pass

        for name, child in self.get_children():
            olist = child.save2(fid, olist)
        return olist

    def save_data2(self, data=None):
        # first data is version, each class file can
        # define version
        if data is None:
            data = {}
        param = {'_status': self._status,
                 '_items':  [x for x in set(self._items)],
                 '_can_have_child':  self._can_have_child,
                 '_has_private_owndir':  self._has_private_owndir,
                 '_suppress':  self._suppress,
                 '_name_readonly':  self._name_readonly,
                 '_genuine_name': self._genuine_name, }
        data['TreeDict'] = (1, param)
        if not self.check_ownitem_exists():
            print("incomplete own items" + str(self))
        return data

    def load(self, fid, olist=None, nlist=None,
             parent=None, keep_zorder=False):
        #
        #  loading from file is a bit triky,
        #  since it appears that you need to have
        #  an object made before loading. Actually,
        #  this routine read contents in file and
        #  put it into a new object. It returns
        #  the new object and self itself will be
        #  discard in the caller.
        #
        if olist is None:
            olist = []
        if nlist is None:
            nlist = []

        try:
            h2 = pickle.load(fid)
        except:
            import traceback
            traceback.print_exc()
            traceback.print_stack()
            return None, olist, nlist

        c = olist.count(h2["id"])
        if c == 0:
            mod = __import__(h2["module"])
            md = sys.modules[h2["module"]]
            the_class = getattr(md, h2["name"])

            # print "loading ", the_class
            obj = the_class(src=fid)
            obj._name = h2["sname"]
            obj._var0 = h2["var0"]
#           if isinstance(obj._var0, PyContents):
#               obj._var0.set_td(obj)
            if "var" in h2:
                for key in h2["var"]:
                    obj._var[key] = h2["var"][key]
            obj._note = h2["note"]
            if 'format' in h2:
                dprint2('Fileformat ', h2["format"])
                if h2["format"] == 1:
                    obj.load_data(fid)
                else:
                    try:
                        data = pickle.load(fid)
                        obj.load_data2(data)
                    except:
                        import traceback
                        traceback.print_exc()
            else:
                obj.load_data(fid)

            olist.append(h2["id"])
            nlist.append(obj.id)
        else:
            idx = olist.index(h2["id"])
            raise ValueError("Treedata is entangled...!?")
            # obj=parent.find_child(id=nlist[idx])

        for i in range(0, h2["num_child"]):
            try:
                child, olist, nlist = TreeDict().load(fid,
                                                      olist=olist, nlist=nlist,
                                                      parent=obj, keep_zorder=keep_zorder)
            except:
                import traceback
                traceback.print_exc()
                #print('File broken !!!!!')
                break

            if child is not None:
                obj.add_child(child, keep_zorder=keep_zorder)
#            if child._name == 'book7': break
#        print 'returning', obj
        return obj, olist, nlist

    def load_data(self, fid=None):
        h2 = pickle.load(fid)

    def load_data2(self, data):
        if not "TreeDict" in data:
            return
        d = data["TreeDict"]
        if len(d) > 1:
            ### this is a ugly patch to fix a bug once exists ###
            if '_items' in d[1]:
                d[1]['_items'] = [x for x in set(d[1]['_items'])]
            ###
            for key in d[1]:
                setattr(self, key, d[1][key])

    def remove_ownitem(self, items=None):
        items = self.ownitem(items)
        for item in items:
            self.rm(item)

    def check_ownitems(self):
        for obj in self.walk_tree():
            for item in obj.ownitem():
                fpath = os.path.join(obj.owndir(), item)
                if not os.path.exists(fpath):
                    print("not found :" + fpath)

    def check_ownitem_exists(self):
        for item in self.ownitem():
            fpath = os.path.join(self.owndir(), item)
            if not os.path.exists(fpath):
                return False
        return True

    def init_after_load(self, olist, nlist):
        pass

    def resolve_olist_nlist_map(self, oid, olist, nlist):
        if olist.count(oid) == 0:
            return None
        idx = olist.index(oid)
        nid = nlist[idx]
        obj = self.find_by_id(nid)
        if obj is not None:
            return weakref.proxy(obj)
        return None

    def copy_tree(self, fid):
        self.save(fid)

    def paste_tree(self, fid):
        obj, ol, nl = TreeDict().load(fid)
        name = obj.get_next_name(obj.get_namebase())
        if obj.name != name:
            print(("renaming object name", obj.name, "->", name))
        self.add_child(name, obj)
        return obj, ol, nl

    def duplicate(self, new_parent, new_name='', save_script_link=False,
                  ignore_name_readonly=True):
        '''
        duplicate self under a new parent with new_name
        duplicate(self, new_parent, new_name, save_script_link=False)
        '''
        from ifigure.ifigure_config import st_scratch
        from ifigure.mto.py_script import PyScript

        name = threading.current_thread().name
        if new_name == '':
            new_name = self.name
#        sc = st_scratch + '_' + str(os.getpid()) + name
        sc = os.path.join(self.get_root_parent().getvar('wdir'),
                          '.duplicate_sc_'+name)

        if save_script_link:
            scripts = [obj for obj in self.walk_tree()
                       if isinstance(obj, PyScript)]
            for s in scripts:
                s._save_link = True
        flag, tmpdir = self.save_subtree(sc, maketar=False)
        if save_script_link:
            for s in scripts:
                s._save_link = False

        child = new_parent.load_subtree(tmpdir, usetar=False)  # st_scratch)
        if child.name != new_name:
            child.rename(new_name, ignore_name_readonly=ignore_name_readonly)

    # utilities to load/save subtree from/to file
    def save_subtree(self, filename, compress=False, maketar=True):
            # save data to temporary directory
        dname = '.###_'+os.path.basename(filename)
        tmpdir = os.path.join(os.path.dirname(filename), dname)
        if os.path.exists(tmpdir):
            m = working_dir_exists_message1(str(tmpdir))
            ret = dialog.message(None,
                                 m,
                                 'Working Directory already exists.',
                                 2)
            if ret == 'ok':
                shutil.rmtree(tmpdir)
            else:
                return False
        os.mkdir(tmpdir)
        if self.has_owndir() and self.get_extfolderpath() is None:
            if self._has_private_owndir:
                #            if self._can_have_child:
                dd = os.path.join(tmpdir, self.name)
                if os.path.exists(dd):
                    shutil.rmtree(dd)
                shutil.copytree(self.owndir(), dd)
            else:
                for item in self.ownitem():
                    #                    print('coping', os.path.join(self.owndir(), item),  os.path.join(tmpdir, item))
                    shutil.copyfile(os.path.join(self.owndir(), item),
                                    os.path.join(tmpdir, item))

        fpath = os.path.join(tmpdir, ".tree_data")
#         dprint1(fpath)
        try:
            fid = open(fpath, 'wb')
            self.save2(fid)
            fid.close()
        except IOError as error:
            print(traceback.format_exc())
            return False
        if not maketar:
            return True, tmpdir

        # make tar or tar.gz file
        if compress:
            mode = 'w:gz'
        else:
            mode = 'w'
        fid = open(filename, 'wb')
        tfid = tarfile.open(mode=mode, fileobj=fid)
        basename = os.path.basename(tmpdir)
        for item in os.listdir(tmpdir):
            dprint1('adding item ' + str(item))
            tfid.add(os.path.join(tmpdir, item),
                     arcname=os.path.join(basename, item))
            dprint1('(done)')
        tfid.close()
        fid.close()

        shutil.rmtree(tmpdir)
        return True

    def load_subtree(self, filename, keep_zorder=True, message=None,
                     compress=False, usetar=True):
        '''
        load contents of filename and
        added as self's child
        '''
        if usetar:
            filenamet = filename
            if compress:
                mode = 'r:gz'
            else:
                mode = 'r'
            fid = open(filenamet, 'rb')
            tfid = tarfile.open(mode=mode, fileobj=fid)
#             print "checking file contents..."

            for member in tfid.getmembers():
                if str(os.path.basename(member.name)) == ".tree_data":
                    org_dir = os.path.dirname(member.name)
                    break
            else:
                print("invalid file")
                tfid.close()
                fid.close()
                return None

#             tar_xpath = os.path.dirname(filenamet)
#             tar_xpath2 = os.path.join(os.path.dirname(filenamet), org_dir)
            wdir = self.get_root_parent().getvar('wdir')
            tar_xpath = wdir
            tar_xpath2 = os.path.join(wdir, org_dir)

            if os.path.exists(tar_xpath2):
                m = working_dir_exists_message1(str(tar_xpath2))
                ret = dialog.message(None,
                                     m,
                                     'Working Directory already exists.',
                                     2)

                if ret == 'ok':
                    shutil.rmtree(tar_xpath2)
                else:
                    tfid.close()
                    fid.close()
                    return None

            tfid.extractall(path=tar_xpath)
            tfid.close()
            fid.close()

        else:
            tar_xpath2 = filename

        fpath = os.path.join(tar_xpath2, ".tree_data")
        fid = open(fpath, 'rb')
        try:
            td, olist, nlist = TreeDict().load(fid, keep_zorder=True)
        except:
            app = wx.GetApp().TopWindow
            dialog.showtraceback(parent=app,
                                 txt='Failed to load subtree',
                                 title='Fail to load',
                                 traceback=traceback.format_exc())

#         print olist, nlist
        fid.close()

        oname = td._name
        names = self.get_childnames()

        top = self.get_root_parent()

        # dirty if switch...;D
        from ifigure.mto.py_code import PyFolder
        from ifigure.mto.py_extfile import PyExtFolder, ExtMixIn
        if isinstance(td, ExtMixIn):
            tmp_obj = PyExtFolder()
        else:
            tmp_obj = PyFolder()

        top.add_child('tmp', tmp_obj, keep_zorder=True)

        imodel = tmp_obj.add_child(oname, td, keep_zorder=keep_zorder)

#         print time.time()-t0
        # move owndir
        if td._has_private_owndir:
            #         if td._can_have_child:
            fpath = os.path.join(tar_xpath2, oname)
            if os.path.exists(fpath):
                #                if not td.has_owndir(): td.mk_owndir()
                #                this may not work if two are not in the same filesystem
                #                os.rename(fpath, td.owndir())
                shutil.move(fpath, td.owndir())
        else:
            if td.has_item():
                td.mk_owndir()
            for item in td.ownitem():
                fpath = os.path.join(tar_xpath2, item)
                if os.path.exists(fpath):
                    shutil.move(fpath,
                                os.path.join(td.owndir(), item))
#         print time.time()-t0

        if names.count(oname) != 0:
            td.rename(self.get_next_name(oname+'_loaded'),
                      ignore_name_readonly=True)

#         if not td._can_have_child and td.has_item(): self.mk_owndir()
        if not td._has_private_owndir and td.has_item():
            self.mk_owndir()

        try:
            td.move(self, keep_zorder=True)
        except:
            traceback.print_exc()
            if message is not None:
                message.txt = (str(td) + ' can not paste under ' +
                               str(self))
            else:
                traceback.print_exc()
            td.destroy()
            tmp_obj.destroy()
            shutil.rmtree(tar_xpath2)
            return None

        tmp_obj.destroy()

        for obj in td.walk_tree(stop_at_ext=True):
            obj.init_after_load(olist, nlist)

        # delete temporary directory
        shutil.rmtree(tar_xpath2)

        if not self.can_have_child(td):
            if message is not None:
                message.txt = (str(td) + ' can not paste under ' +
                               str(self))
            td.destroy()
            return None
#         print time.time()-t0
        return td

    def random_tmp_name(self, seed='1'):
        # generate a random string for an initial work dir
        from datetime import datetime
        import hashlib
        strtime = datetime.now().strftime("%Y-%m-%d %H:%M:%S. %f")
        # print strtime
        m = hashlib.md5()
        bytechar = (strtime+str(seed)).encode('latin-1')
        m.update(bytechar)
        txt = m.hexdigest()
        return txt

    def has_item(self):
        return len(self._items) != 0

    def ownitem(self, items=None):
        '''
        return ownitems' path. 
        Without argumetns, it returns the list of paths.
        With argument of items it returns paths only specified by
        the argument. 
        Note, if getvar returns None, it is excluded
        '''
        if items is None:
            return [self.getvar(x) for x in self._items if self.getvar(x) is not None]
        else:
            return [self.getvar(x) for x in items
                    if self.getvar(x) is not None and x in self._items]

    def prepare_compact_savemode(self):
        return self._var

    def sort_children_up(self):
        names = self.get_childnames()
        for i, name in enumerate(sorted(names)):
            oi = self.i_child(name=name)
            self.move_child(oi, i)

    def sort_children_down(self):
        names = self.get_childnames()
        names = [i for i in reversed(sorted(names))]
        for i, name in enumerate(names):
            oi = self.i_child(name=name)
            self.move_child(oi, i)

    def sort_var_up(self):
        self._var = collections.OrderedDict(sorted([(k, self._var[k])
                                                    for k in self._var]))

    def sort_var_down(self):
        self._var = collections.OrderedDict(reversed(sorted([(k, self._var[k])
                                                             for k in self._var])))

    def search(self, txt, dir=0, include_self=True):
        start = False
        if dir == 0:
            g = self.get_root_parent().walk_tree
        else:
            g = self.get_root_parent().walk_tree_rev
        for x in g():
            if x is self:
                start = True
                if not include_self:
                    continue
            if not start:
                continue
            if x._name.upper().find(txt.upper()) != -1:
                return x
        return None

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        return self._name

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, value):
        self._status = value

    def get_auto_status_str(self):
        return ''


class TopTreeDict(TreeDict):

    #   This class defines the capability which only tree
    #   top should have
    #      Event handling for
    #         Save/Load, Cut/Paste

    def __init__(self, src=None):
        self._focus = None
        self._app = None   # proxy to app
        super(TopTreeDict, self).__init__()
        self._name_readonly = True
        self.setvar("filename", None)
        self.setnote("filename", "name of file")

        cond = True
        while cond:
            if src is not None:
                return
            txt = self.random_tmp_name()

#           wdir=os.path.join(ifigure_config.rcdir,
#                     '###untitled_'+txt)
            wdir = os.path.join(ifigure_config.tempdir,
                                '###untitled_'+txt)

            #print("preparing working dir : checking " + wdir)
            time.sleep(0.5)
            if os.path.exists(wdir) == False:
                os.mkdir(wdir)
                os.mkdir(os.path.join(wdir, '.trash'))
                cond = False
        self.setvar("wdir", wdir)
        self.setnote("wdir", "working directory")

    @classmethod
    def isTopTreeDict(self):
        return True

    def get_namebase(self):
        return 'toptreedict'

    def tree_viewer_menu(self):
        # return MenuString, Handler, MenuImage
        return None

    def get_trash(self):
        wdir = self.getvar("wdir")
        return os.path.join(wdir, '.trash')

    def set_app(self, app):
        self._app = weakref.proxy(app)

    @property
    def app(self):
        return self._app

#    def GetFocus(self):
#        pass
#    def SetFocus(self):
#        pass

    def onCut(self, e):
        pass

    def onPaste(self, e):
        pass

    def onCopy(self, e):
        pass

    def onProjTreeActivate(self, e):
        pass

    def SaveToFile(self, filename, opath=None):
        # save data to temporary directory
        old_wdir = self.getvar("wdir")
        old_filename = self.getvar("filename")

        self._set_work_dir(filename)
        d = self.getvar("wdir")

        if old_wdir != d:
            if os.path.exists(d):
                shutil.rmtree(d)
            shutil.copytree(old_wdir, d)

        fpath = os.path.join(d, ".tree_data")
        try:
            fid = open(fpath, 'wb')
            #  this save should include
            #  writing all files including on-memory scripts
            #  to files.
            self.save(fid)
            fid.close()
        except IOError as error:
            return False

        # set proj.filename
        self.setvar("filename", filename)
#         if old_wdir != d:
#              self._delete_tempdir(old_wdir)
        fid = open(os.path.join(d, '.filename'), 'w')
        fid.write(filename)
        fid.close()
        print("done....(save)")

        from ifigure.utils.mp_tarzip import MPTarzip
        success = MPTarzip().Run(filename, d, old_wdir)

        if not success:
            self.setvar("filename", old_filename)
            self.getvar("wdir", old_wdir)            

        return success

    def LoadFromFile(self, filename, sb=None):
        if sb is not None:
            sb.SetStatusText("loading project file " + filename, 0)
            sb.Refresh()
        fid = open(filename, 'rb')
        tfid = tarfile.open(mode='r:gz', fileobj=fid)
#         sys.stdout.write("checking file contents...")
        print("checking file contents...")
        for member in tfid.getmembers():
            if str(os.path.basename(member.name)) == ".tree_data":
                org_dir = os.path.dirname(member.name)
                break
        else:
            sys.stdout.write("invalid file\n")
#             print("invalid file")
            tfid.close()
            fid.close()
            return None

        tar_xpath2 = self._work_dirname(filename)
        tar_xpath = os.path.join(tar_xpath2, org_dir)
#         tar_xpath2 = os.path.join(os.path.dirname(filename), org_dir)

        do_expand = True

        crash_file = self._check_crash_file(filename)
        if crash_file != '':
            m = working_dir_exists_message2(str(crash_file))
            ret = dialog.message(None, m, '', 5)
            if ret == 'yes':
                shutil.rmtree(crash_file)
                if len(os.listdir(os.path.dirname(crash_file))) == 0:
                    #                  dprint1('deleting', os.path.dirname(crash_file))
                    shutil.rmtree(os.path.dirname(crash_file))
            elif ret == 'no':
                tfid.close()
                fid.close()
                trash = os.path.join(crash_file, '.trash')
                if os.path.exists(trash):
                    shutil.rmtree(trash)
#              dprint1('moving', crash_file, tar_xpath2)
                shutil.move(crash_file, tar_xpath2)
                if len(os.listdir(os.path.dirname(crash_file))) == 0:
                    #                 dprint1('deleting', os.path.dirname(crash_file))
                    shutil.rmtree(os.path.dirname(crash_file))
                do_expand = False
            else:
                tfid.close()
                fid.close()
                return
        if do_expand:
            os.mkdir(tar_xpath2)
            # print 'expanding to', tar_xpath2
            tfid.extractall(path=tar_xpath2)
            tfid.close()
            fid.close()
            for item in os.listdir(tar_xpath):
                src = os.path.join(tar_xpath, item)
                dst = os.path.join(tar_xpath2, item)
                shutil.move(src, dst)
            shutil.rmtree(tar_xpath)

        fpath = os.path.join(tar_xpath2, ".tree_data")
        fid = open(fpath, 'rb')

        real_top, olist, nlist = TreeDict().load(fid, keep_zorder=True)
#         print olist, nlist
        fid.close()
#         sys.stdout.write("done (load)\n")
        print("done....(load)")

        self.CloseProject()

        real_top.setvar("filename", str(filename))
        real_top.setvar("wdir", str(tar_xpath2))
        real_top._app = self._app
        os.mkdir(os.path.join(tar_xpath2, '.trash'))
        fid = open(os.path.join(tar_xpath2, '.filename'), 'w')
        fid.write(filename)
        fid.close()

        if sb is not None:
            sb.SetStatusText("initializing model :"+filename, 0)
            sb.Refresh()
        print("initializing model...")
        for obj in real_top.walk_tree(stop_at_ext=True):
            obj.init_after_load(olist, nlist)
        print("done....(init) " + str(real_top))

        return real_top

    def CloseProject(self):
        path = self.getvar("filename")
        wdir = self.getvar("wdir")
        if wdir is not None:
            self._delete_tempdir(wdir)
        self.destroy()

    def _check_crash_file(self, file):
        from ifigure.ifigure_config import tempdir_base
        from ifigure.utils.pid_exists import pid_exists

        crashed_process = []
        for item in os.listdir(tempdir_base):
            if item.startswith('pid'):
                pid = int(item[3:])
                if not pid_exists(pid):
                    #       try:
                    #          os.kill(pid, 0)
                    #       except OSError:
                    dname = os.path.join(tempdir_base, item)
                    for item2 in os.listdir(dname):
                        f = os.path.join(tempdir_base, item,
                                         item2, '.filename')
                        if os.path.exists(f):
                            fid = open(f, 'r')
                            l = fid.readline()
                            fid.close()
                            if str(l) == str(file):
                                return os.path.join(tempdir_base, item, item2)
        return ''

    def _set_work_dir(self, path):
        d = self._work_dirname(path)
        self.setvar("wdir", d)

    def _work_dirname(self, path):
        #        base=os.path.dirname(path)
        #        dname='.###ifigure_'+os.path.basename(path)
        #        d=os.path.join(base, dname)
        dname = '.###ifigure' + \
            '_'.join((os.path.splitdrive(str(path))[1]).split(os.sep))
        d = os.path.join(ifigure_config.tempdir, dname)
        return d

    def _delete_tempdir(self, old_wdir):
        #print("deleting temporary dir " + old_wdir)
        try:
            shutil.rmtree(old_wdir)
        except:
            print('failed to delete temporary directory')
            import __main__
            __main__.xxx.append(old_wdir)
            pass
        # os.rmdir(old_wdir)

    def memorized_data(self):
        '''
        list of memorized data when tree_obj is
        destoried.
        '''
        return []


#
#   example
#
if __name__ == '__main__':
    print("Demonstration")
    ##
    # Point of this class
    ##
    ##     children is stored in OrderedDict
    # adder need to be implemented to add a
    # child

    root = TreeDict()
    root.add_child("test", TreeDict())
    root.add_child("test2", TreeDict())
    root.test.add_child("test3", TreeDict())

    print(("root", root))
    print(("root.test", root.test))
    print(("root.test.test3", root.test.test3))
    print(("root.test2", root.test2))

    gen = root.walk_tree()
    print(gen.next().get_full_path())
    print(gen.next().get_full_path())
    print(gen.next().get_full_path())
    print(gen.next().get_full_path())
    print(gen.next().get_full_path())


if __name__ == '__main__':
    print("Demonstration")
    ##
    # Point of this class
    ##
    ##     children is stored in OrderedDict
    # adder need to be implemented to add a
    # child

    page = FigPage()
    page.add_axes('axes_1')
    page.add_axes('axes_2')
    page.axes1.add_plot()

    # or you can add this way
    plot = FigPlot()
    page.axes1.add_plot("plot_1", plot)

    # in this case you need to apply figobj

    page.axes1.add_figobj("plot2")

    #  attribute setting
    page.setp('area', [0, 1, 0, 1])
    print(page.getp("area"))

    # if you do  getp_all, pay attention that
    # the returned value is the SAME object
    att = page.getp_all()
    print(att)
    att["area"] = [0, 0, 0, 0]
    a = page.getp("area")
    print(a)

    ### this generate tree list ###
    print(page.list_all())

    # can access read-only member
    # if getter is properly set
    print(page.axes1.plot.id)

    #   page.axes1.plot.id=100
    #   this will cause error (read-only)

    print(page.axes1.plot)

    ### this is not allowed
    #  in other word, all members are read-only
    #  page.axes1.plot.id2=3
    #  page.axes1.plot.test.id2=3
    #
    # you can still do this but are not supposed to do
    #  Do not create "_" private member.
    #  page.axes1._test=3
    # if you did not the above, the nexe will cause error
    # print page.axes1._test

   #   print dir(page.axes1.plot)



import socket
import os
import subprocess


def PickUnusedPort():
    '''
    a routeint to le a system to pick an unused port.
    this recipe was found in active state web site
    '''

    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(('localhost', 0))
    addr, port = s.getsockname()
    s.close()
    return port


if __name__ == '__main__':
    port = PickUnusedPort()
    hostname = os.getenv("HOSTNAME")
    command = 'python -m Pyro4.naming -p '+str(port) + ' -n '+hostname + ' &'
    subprocess.Popen(command, shell=True)



from __future__ import print_function
#
#   fig_image
#
#  History:
#          12.06.10  Added Highlight
#          12.06.11  Added custom picker
#          12.09.04  Added image interpolation
#
#  Author :
#         Syun'ichi Shiraiwa
#  E-mail :
#         shiraiwa@psfc.mit.edu
#
# *******************************************
#     Copyright(c) 2012- PSFC, MIT
# *******************************************

from ifigure.ifigure_config import isMPL2
from ifigure.mto.fig_obj import FigObj, mask_negative
from ifigure.mto.axis_user import XUser, YUser, CUser, ZUser
from ifigure.widgets.canvas.file_structure import *
import ifigure
import os
import sys
import numpy as np
import ifigure.utils.cbook as cbook
import ifigure.widgets.canvas.custom_picker as cpicker
from ifigure.utils.interpolate import interp2d
from scipy.interpolate import griddata
from ifigure.utils.cbook import ProcessKeywords
from ifigure.utils.triangulation_wrapper import tri_args
from matplotlib.cm import ScalarMappable
from matplotlib.patches import Rectangle
from matplotlib.transforms import Bbox
from ifigure.utils.args_parser import ArgsParser
from matplotlib.colors import Colormap

#
#  debug setting
#
import ifigure.utils.debug as debug
debug.debug_default_level = 1
dprint1, dprint2, dprint3 = debug.init_dprints('FigImage')

default_kargs = {'use_tri': False,
                 #                 'interp' : 'linear',
                 'interp': 'nearest',
                 'shading': 'flat',
                 'alpha':  None}
#                 'cmap'   :  'jet'}


class FigImage(FigObj, XUser, YUser, ZUser, CUser):
    """
    image : show image 

    image(z)
    image(x, y, z)
    """
    default_rasterized = True

    def __new__(cls, *args, **kywds):
        def set_hidden_vars(obj):
            if not hasattr(obj, '_tri'):
                obj._tri = None  # this can go away!?
            for key in default_kargs:
                if not obj.hasp(key):
                    obj.setp(key, default_kargs[key])
                if not obj.hasvar(key):
                    obj.setvar(key, default_kargs[key])
            return obj

        if 'src' in kywds:
            obj = FigObj.__new__(cls, *args, **kywds)
            obj = set_hidden_vars(obj)
            return obj

        p = ArgsParser()
        p.add_opt('x', None, ['numbers|nonstr', 'dynamic'])
        p.add_opt('y', None, ['numbers|nonstr', 'dynamic'])
        p.add_var('z', ['numbers|nonstr', 'dynamic'])
        p.set_pair("x", "y")  # x and y should be given
        # together
        p.set_ndconvert("x", "y", "z")
        p.set_squeeze_minimum_1D("x", "y", "z")
        p.set_default_list(default_kargs)
        p.add_key2(('use_tri', 'interp', 'shading', 'alpha'))
        p.add_key('cmap', 'jet')

        v, kywds, d, flag = p.process(*args, **kywds)

        if not flag:
            raise ValueError('Failed when processing argument')

        obj = FigObj.__new__(cls, *args, **kywds)
        obj = set_hidden_vars(obj)

        for name in ('x', 'y', 'z', 'use_tri', 'interp', 'shading', 'alpha'):
            obj.setvar(name, v[name])
        if v['cmap'] is not None:
            if isinstance(v['cmap'], Colormap):
                v['cmap'] = v['cmap'].name
            kywds['cmap'] = v['cmap']
            del v['cmap']
        obj.setvar("kywds", kywds)

        return obj

    def __init__(self, *args, **kywds):
        self._data_extent = None
        XUser.__init__(self)
        YUser.__init__(self)
        ZUser.__init__(self)
        CUser.__init__(self)

        self._pick_pos = None
        self._cb_added = False
        args = []
        if 'src' not in kywds:
            kywds = self.getvar("kywds")
        super(FigImage, self).__init__(**kywds)

    @classmethod
    def isFigImage(self):
        return True

    @classmethod
    def can_have_child(self, child=None):
        return False

    @classmethod
    def get_namebase(self):
        return 'image'

    @classmethod
    def property_in_file(self):
        return (["array", "extent"] +
                super(FigImage, self).property_in_file())

    @classmethod
    def property_in_palette(self):
        return (['image'], [["image_interp", "alpha_2", "caxis", "noclip3d"], ])

    @classmethod  # define _attr values to be saved
    def attr_in_file(self):
        return (["interp", "shading", "alpha"] +
                super(FigImage, self).attr_in_file())

    @classmethod
    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx1 = cbook.LoadImageFile(path, 'image.png')
        return [idx1]

    def set_parent(self, parent):
        XUser.unset_ax(self)
        YUser.unset_ay(self)
        ZUser.unset_az(self)
        CUser.unset_ac(self)
        super(FigImage, self).set_parent(parent)
        XUser.get_xaxisparam(self)
        YUser.get_yaxisparam(self)
        ZUser.get_zaxisparam(self)
        CUser.get_caxisparam(self)

    def args2var(self):
        ret = self._args2var()
        if ret:
            if 'cmap' in self.getvar("kywds"):
                cax = self.get_caxisparam()
                cax.set_cmap(self.getvar('kywds')['cmap'])
        return ret

    def _args2var(self):
        names0 = self.attr_in_file()
        names = ["x", "y", "z"] + names0
        use_np = [True]*3 + [False]*len(names0)
        values = self.put_args2var(names,
                                   use_np)
        x = values[0]
        y = values[1]
        z = values[2]
        if y is None and x is None and z.ndim == 2:
            y = np.arange((z.shape)[-2]).astype(z.dtype)
            self.setp("y", y)
            x = np.arange((z.shape)[-1]).astype(z.dtype)
            self.setp("x", x)
        if x is None:
            return False
        if y is None:
            return False
        if self._tri is not None:
            if self._tri.shape[0] == z.size:
                self.setp("x", x)
                self.setp("y", y)
                self.setp("z", z)
                return True
        if (x.size*y.size != z.size and
                not (x.size == z.size and y.size == z.size)):
            self.setp("x", None)
            self.setp("y", None)
            self.setp("z", None)
            return False
        return True

    def generate_artist(self):
        # this method generate artist
        # if artist does exist, update artist
        # based on the information specifically
        # managed by fig_obj tree. Any property
        # internally managed by matplotlib
        # does not change
        container = self.get_container()
        if container is None:
            return
#           if self.get_figaxes().get_3d(): return
        if self.isempty() is False:
            return
        x, y, z = self._eval_xyz()
        if z is None:
            return

        lp = self.getp("loaded_property")
        aspect = container.get_aspect()
        cax = self.get_caxisparam()
        if cax is None:
            dprint1('Error: cax is None')
            return
        crange = cax.range

        if lp is None or len(lp) == 0:
            if (self._tri is not None and
                    self._tri.shape[0] == z.size):
                pass

            else:
                if (x.size*y.size != z.size and
                        not (x.size == z.size and y.size == z.size)):
                    print('FigImage: array size is wrong')
                    return

            if not self.getvar('use_tri'):
                if self.get_figaxes().get_3d():
                    xp, yp, zp = x, y, z
                    extent = (np.min(x), np.max(x),
                              np.min(y), np.max(y), )
                else:
                    xp, yp, zp = self.interp_image(x, y, z)
                    extent = (np.min(xp), np.max(xp),
                              np.min(yp), np.max(yp), )
                args = []
                kywds = self._var["kywds"].copy()
                kywds['alpha'] = self.getp('alpha')
    #              args.append(np.flipud(zp))

                if cax.scale == 'linear':
                    args.append(zp)
                    kywds["vmin"] = crange[0]
                    kywds["vmax"] = crange[1]
                else:
                    # args.append(np.log10(zp))
                    args.append(zp)
                    kywds["vmin"] = np.log10(max((crange[0], 1e-16)))
                    kywds["vmax"] = np.log10(max((crange[1], 1e-16)))

                kywds["aspect"] = aspect
                kywds["origin"] = 'lower'
                if self.getp("interp") == 'linear':
                    kywds["interpolation"] = 'bilinear'
                elif self.getp("interp") == 'cubic':
                    kywds["interpolation"] = 'bicubic'
                else:
                    kywds["interpolation"] = self.getp("interp")

                self.set_artist(container.imshow(*args,
                                                 # picker=cpicker.Picker,
                                                 extent=extent,  **kywds))
                cax.set_crangeparam_to_artist(self._artists[0])

                setattr(self._artists[0].get_array(), '_xyp', (xp, yp))
            else:
                #print('drawing tri image')
                if x.size*y.size == z.size:
                    x, y = np.meshgrid(x, y)

                args, self._tri = tri_args(x, y, self._tri)
                kywds = self._var["kywds"]
                kywds['alpha'] = self.getp('alpha')
                if cax.scale == 'linear':
                    # args.append(z.flatten().astype(float))
                    args.append(np.real(z.flatten()))
                    kywds["clim"] = (crange[0], crange[1])
                else:
                    # args.append(np.log10(z))
                    # args.append(z.flatten().astype(float))
                    args.append(np.real(z.flatten()))
                    kywds["clim"] = [np.log10(max((crange[0], 1e-16))),
                                     np.log10(max((crange[1], 1e-16)))]
                kywds['shading'] = self.getp('shading')
                kywds['mask'] = self.getp('mask')

                self.set_artist(container.tripcolor(*args,
                                                    **kywds))
                cax.set_crangeparam_to_artist(self._artists[0])
            self._data_extent = [np.min(x), np.max(x),
                                 np.min(y), np.max(y)]
        else:
            if self.getvar('use_tri'):
                #print('redrawing tri image')
                x, y, z = self.getp(('x', 'y', 'z'))
                args, self._tri = tri_args(x, y, self._tri)

                args.append(z.flatten().astype(float))
                # astype(float) is patch work to open past files..

                #kywds = self._var["kywds"]
                kywds = lp[0]
                kywds['shading'] = self.getp('shading')
                kywds['mask'] = self.getp('mask')
                kywds['alpha'] = self.getp('alpha')
                keys = ['alpha', 'cmap', 'zorder']
                for k in keys:
                    if k in lp[0]:
                        kywds[k] = lp[0][k]
#                  print lp
#                  kywds["clim"]=lp[0]["clim"]
                del(kywds['array'])
                self.set_artist(container.tripcolor(*args,
                                                    **kywds))
                cax.set_crangeparam_to_artist(self._artists[0])

            else:
                x, y, z = self.getp(('x', 'y', 'z'))
                xp, yp, zp = self.interp_image(x, y, z)
                args = []
                kywds = {}
                if 'im_center' in self._var["kywds"]:
                    kywds['im_center'] = self._var["kywds"]['im_center']
                if 'im_axes' in self._var["kywds"]:
                    kywds['im_axes'] = self._var["kywds"]['im_axes']

                kywds['alpha'] = self.getp('alpha')
                if self.getp("interp") == 'linear':
                    kywds["interpolation"] = 'bilinear'
                elif self.getp("interp") == 'cubic':
                    kywds["interpolation"] = 'bicubic'
                else:
                    kywds["interpolation"] = self.getp("interp")

                if cax.scale == 'linear':
                    args.append(zp)
                    kywds["vmin"] = crange[0]
                    kywds["vmax"] = crange[1]
                else:
                    args.append(zp)
                    kywds["vmin"] = np.log10(max((crange[0], 1e-16)))
                    kywds["vmax"] = np.log10(max((crange[1], 1e-16)))
                keys = ['alpha', 'cmap', 'zorder']
                for k in keys:
                    if k in lp[0]:
                        kywds[k] = lp[0][k]
                self.set_artist(container.imshow(*args,
                                                 extent=lp[0]["extent"], aspect=aspect,
                                                 origin='lower', **kywds))
                cax.set_crangeparam_to_artist(self._artists[0])
                setattr(self._artists[0].get_array(), '_xyp', (xp, yp))
        self.delp("loaded_property")

        self.set_rasterized()
        # this is to resize image when window is resized...
        if not self._cb_added:
            fig_page = self.get_figpage()
            fig_page.add_resize_cb(self)
            self._cb_added = True

#           for artist in self._artists:
#               artist.figobj=self
#               artist.figobj_hl=[]
#               artist.set_zorder(self.getp('zorder'))
    def onResize(self, evt):
        self.set_interp(self.get_interp(self._artists[0]), self._artists[0])
        self.set_bmp_update(False)

    def del_artist(self, artist=None, delall=False):
        if delall:
            artistlist = self._artists
        else:
            artistlist = self._artists

        self.store_loaded_property()

#        self.highlight_artist(False, artistlist)
        for a in artistlist:
            a.remove()

        super(FigImage, self).del_artist(artistlist)

    def get_mappable(self):
        return [a for a in self._artists if isinstance(a, ScalarMappable)]

    def highlight_artist(self, val, artist=None):
        #        print val, artist
        figure = self.get_figpage()._artists[0]
        ax = self.get_figaxes()
        if artist is None:
            alist = self._artists
        else:
            alist = artist
        if val == True:
            container = self.get_container()
            if container is None:
                return

            de = self.get_data_extent()
            from ifigure.matplotlib_mod.art3d_gl import AxesImageGL
            if isinstance(alist[0], AxesImageGL):
                hl = alist[0].add_hl_mask()
                for item in hl:
                    alist[0].figobj_hl.append(item)

#               hl = alist[0].make_hl_artist(container)
#               rect_alpha = 0.0
            else:
                x = [de[0], de[1], de[1], de[0], de[0]]
                y = [de[2], de[2], de[3], de[3], de[2]]
                hl = container.plot(x, y, marker='s',
                                    color='k', linestyle='None',
                                    markerfacecolor='None',
                                    markeredgewidth=0.5,
                                    scalex=False, scaley=False)
                rect_alpha = 0.3

                hlp = Rectangle((de[0], de[2]),
                                de[1]-de[0],
                                de[3]-de[2],
                                alpha=rect_alpha, facecolor='k',
                                figure=figure,
                                transform=container.transData)
                if ax is not None:
                    x0, y0 = ax._artists[0].transAxes.transform((0, 0))
                    x1, y1 = ax._artists[0].transAxes.transform((1, 1))
                    bbox = Bbox([[x0, y0], [x1, y1]])
                    hlp.set_clip_box(bbox)
                    hlp.set_clip_on(True)

                figure.patches.append(hlp)
                for item in (hl[0], hlp):
                    alist[0].figobj_hl.append(item)
        else:
            for a in alist:
                if len(a.figobj_hl) != 0:
                    a.figobj_hl[0].remove()
                    figure.patches.remove(a.figobj_hl[1])
                a.figobj_hl = []
#
#   Setter/Getter
#

    def set_cmap(self, value, a):
        ca = self.get_caxisparam()
        ca.set_cmap(value)
        ca.set_crangeparam_to_artist(a)
        if self.has_cbar():
            ca.update_cb()

    def get_cmap(self, a=None):
        ca = self.get_caxisparam()
        return ca.cmap
#
#   HitTest
#

    def get_artist_extent(self, a):
        '''
        retrun the extent of artist in device 
        coordinate

        AxesImage object does not have get_window_extent
        '''
        de = self.get_data_extent()
        x1, y1 = self._artists[0].axes.transData.transform((de[0], de[2]))
        x2, y2 = self._artists[0].axes.transData.transform((de[1], de[3]))
        return [x1, x2, y1, y2]

    def picker_a(self, artist, evt):
        from ifigure.matplotlib_mod.art3d_gl import AxesImageGL
        if (isinstance(artist, AxesImageGL) or
                self.getvar('use_tri')):
            hit, extra = artist.contains(evt)
            if hit:
                self._pick_pos = [evt.xdata, evt.ydata]
                return True,  {'child_artist': artist}
            else:
                return False, {}

        # (an old routine when somehow the above did not work)
        # for mpl1.5, normal image plot needs to do this.
        ax = self.get_figaxes()
        hit, extra = ax._artists[0].contains(evt)
        if not hit:
            return False, {}

        de = self.get_data_extent()
        if (evt.xdata > de[0] and
            evt.xdata < de[1] and
            evt.ydata > de[2] and
                evt.ydata < de[3]):
            self._pick_pos = [evt.xdata, evt.ydata]
            return True, {'child_artist': artist}
        self._pick_pos = None
        return False, {}

    def picker_a0(self, artist, evt):
        hit, extra = self.picker_a(artist, evt)
        if hit:
            return hit, extra, 'area', 3
        else:
            return False, {}, None, 0

#
#  Popup in ProjViewer
#
    def tree_viewer_menu(self):
     # return MenuString, Handler, MenuImag
        ac = self.get_caxisparam()
        if ac is None:
            return super(FigImage, self).tree_viewer_menu()
        if ac.cbar_shown():
            m = [('Hide ColorBar', self.onHideCB, None), ]
        else:
            m = [('Show ColorBar', self.onShowCB, None), ]
        return m+super(FigImage, self).tree_viewer_menu()

    def onShowCB(self, e):
        new_cb = False
        if self.has_cbar():
            new_cb = True
        self.show_cbar()
        if new_cb:
            ifigure.events.SendPVAddFigobj(self.get_figaxes())
        ifigure.events.SendPVDrawRequest(self, w=None)
        e.Skip()

    def onHideCB(self, e):
        self.hide_cbar()
        ifigure.events.SendPVDeleteFigobj(self.get_figaxes())
        ifigure.events.SendPVDrawRequest(self, w=None)
        e.Skip()
#
#  Popup in Canvas
#

    def canvas_menu(self):
        ac = self.get_caxisparam()
        if ac is None:
            return super(FigImage, self).tree_viewer_menu()
        if ac.cbar_shown():
            m = [('Hide ColorBar', self.onHideCB1, None), ]
        else:
            m = [('Show ColorBar', self.onShowCB1, None), ]
        menus = [("Show Slice",  self.onSlice, None)] + m + \
            super(FigImage, self).canvas_menu()
        return menus

    def get_slice(self, xin, yin, a=None):
        if a is None:
            a = self._artists[0]
        array = a.get_array()
        axes = a.axes

        x0, y0 = axes.transAxes.transform([0, 0])
        x1, y1 = axes.transAxes.transform([1, 1])

        ix = float(xin-x0)/float(x1-x0)*array.shape[1]
        iy = float(yin-y0)/float(y1-y0)*array.shape[0]

        try:
            zp1 = array[:, int(ix)]
            zp2 = array[int(iy), :]
        except:
            return None, None

#        atrans = axes.transData.transform
#        idtrans = axes.transAxes.inverted().transform
#        print xin, yin
#        print atrans((xin, yin))
#        print idtrans(atrans((xin, yin)))
#        xn, yn = idtrans(atrans((xin, yin)))
#        print array.shape

#        zp1 = array[:, long(xn*array.shape[1])]
#        zp2 = array[long(yn*array.shape[0]), :]

        xp, yp = getattr(array, '_xyp')
        return (yp, np.array(zp1)), (xp, np.array(zp2))

    def onSlice(self, event):
        from ifigure.interactive import figure, plot, nsec, isec, update, title, xlabel, ylabel
        if event.mpl_xydata[0] is None:
            return

        for a in self._artists:
            axes = a.axes
            if axes is None:
                return
            data1, data2 = self.get_slice(event.mpl_xy[0],
                                          event.mpl_xy[1], a)
            if data1 is None:
                continue
            if data2 is None:
                continue
            figure()
            nsec(2)
            ou = update()
            isec(0)
            plot(data1[0], data1[1])
            title('y slice : y = '+str(event.mpl_xydata[1]))
            xlabel('x')
            isec(1)
            plot(data2[0], data2[1])
            title('x slice : x = '+str(event.mpl_xydata[0]))
            xlabel('y')
            update(ou)

    def get_export_val(self, a):
        x, y, z = self._eval_xyz()
        # althouth internally it is called zdata, export will name it cdata,
        # which is more appropriate.
        return {"cdata": z,
                "xdata": x,
                "ydata": y}

    def interp_image(self, x, y, z):

        if isMPL2:
            return x, y, z
        else:
            axes = self.get_container()

            atrans = axes.transAxes.transform
            idtrans = axes.transData.inverted().transform
            p0, p1 = atrans([(0, 0), (1, 1)])
            dx = np.floor(p1[0]-p0[0])+2
            dy = np.floor(p1[1]-p0[1])+2
            xp = idtrans(
                np.transpose(
                    np.vstack((np.floor(p0[0])+np.arange(int(dx)),
                               np.linspace(p0[0], p1[0], dx)))))[:, 0]
            yp = idtrans(
                np.transpose(
                    np.vstack((np.floor(p0[1])-1+np.zeros(int(dy)),
                               np.linspace(p0[1], p1[1], dy)))))[:, 1]

            # eliminate points outside the data range
            #xp = np.array([tmp for tmp in xp if (tmp > np.min(x) and tmp < np.max(x))])
            #yp = np.array([tmp for tmp in yp if (tmp > np.min(y) and tmp < np.max(y))])

            interp = self.getp("interp")
            from scipy.interpolate import RegularGridInterpolator
            if ((x.size*y.size == z.size and interp == 'nearest') or
                    (x.size*y.size == z.size and interp == 'linear')):
                f = RegularGridInterpolator((y, x), z, method=interp,
                                            bounds_error=False, fill_value=np.nan)
                XP, YP = np.meshgrid(xp, yp)
                p1 = np.transpose(np.vstack((YP.flatten(), XP.flatten())))
                zp = f(p1)
                zp = zp.reshape((len(yp), len(xp))).astype(float)
            elif (x.size*y.size != z.size or interp == 'nearest' or
                  np.any(np.isnan(z))):
                if interp == 'quintic':
                    interp = 'cubic'
                X, Y = np.meshgrid(x, y)
                p1 = np.transpose(np.vstack((X.flatten(), Y.flatten())))
                XP, YP = np.meshgrid(xp, yp)
                interp = self.getp("interp")
                # print 'griddata', interp
                zp = griddata(p1, z.flatten(),
                              (XP.flatten(), YP.flatten()),
                              method=str(interp))
                zp = zp.reshape((len(yp), len(xp))).astype(float)

            else:
                f = interp2d(x, y, z, kind=interp)
                zp = f(xp, yp)
    #           print(zp)
    #        if interp == 'nearest':
            # this hide the outside of image when even 'nearest' is
            # used
            zp[:, xp < np.min(x)] = np.nan
            zp[:, xp > np.max(x)] = np.nan
            zp[yp < np.min(y), :] = np.nan
            zp[yp > np.max(y), :] = np.nan

            return xp, yp, zp

    def get_data_extent(self):
        if self._data_extent is not None:
            return self._data_extent

        x, y, z = self._eval_xyz()
        if z is None:
            return self._data_extent
        if x is None or y is None:
            #  x and y may be None if it failed to
            #  evaluable expression
            if z.ndim != 2:
                return self._data_extent
            if x is None:
                x = [0, z.shape[1]]
            if y is None:
                y = [0, z.shape[0]]
            self._data_extent = [min(x), max(x), min(y), max(y)]
        else:
            self._data_extent = [np.min(x), np.max(x),
                                 np.min(y), np.max(y)]

        return self._data_extent

#
#   range
#
    def _get_3d_placement(self):
        if 'im_center' in self._var["kywds"]:
            im_center = np.array(self._var["kywds"]['im_center'])
        else:
            im_center = np.array([0, 0, 0])
        if 'im_axes' in self._var["kywds"]:
            ax1 = np.array(self._var["kywds"]['im_axes'][0])
            ax2 = np.array(self._var["kywds"]['im_axes'][1])
        else:
            ax1 = np.array([1, 0, 0])
            ax2 = np.array([1, 0, 0])

        return im_center, ax1, ax2

    def get_xrange(self, xrange=[None, None], scale='linear'):
        x, y, z = self._eval_xyz()
        if x is None:
            return xrange
        if not self.get_figaxes().get_3d():
            if scale == 'log':
                x = mask_negative(x)
            return self._update_range(xrange, [np.nanmin(x), np.nanmax(x)])
        else:
            im_center, ax1, ax2 = self._get_3d_placement()
            data = np.hstack(
                [im_center[0] + ax1[0]*x, im_center[0] + ax2[0]*y])
            xrange = self._update_range(xrange,
                                        (np.nanmin(data), np.nanmax(data)))
            return xrange

    def get_yrange(self, yrange=[None, None],
                   xrange=[None, None], scale='linear'):
        #        de = self.get_data_extent()
        x, y, z = self._eval_xyz()
        if x is None:
            return yrange
        if y is None:
            return yrange
        if not self.get_figaxes().get_3d():
            if scale == 'log':
                y = mask_negative(y)
            return self._update_range(yrange, (np.nanmin(y), np.nanmax(y)))
        else:
            im_center, ax1, ax2 = self._get_3d_placement()
            data = np.hstack(
                [im_center[1] + ax1[1]*x, im_center[1] + ax2[1]*y])
            yrange = self._update_range(yrange,
                                        (np.nanmin(data), np.nanmax(data)))
            return yrange

    def get_zrange(self, zrange=[None, None],
                   xrange=[None, None],
                   yrange=[None, None],
                   scale='linear'):
        if not self.get_figaxes().get_3d():
            return zrange

        x, y, z = self._eval_xyz()
        im_center, ax1, ax2 = self._get_3d_placement()

        data = np.hstack([im_center[2] + ax1[2]*x, im_center[2] + ax2[2]*y])
        zrange = self._update_range(zrange,
                                    (np.nanmin(data), np.nanmax(data)))

        return zrange

    def get_crange(self, crange=[None, None],
                   xrange=[None, None],
                   yrange=[None, None],
                   scale='linear'):

        x, y, z = self._eval_xyz()

        if np.iscomplexobj(z):
            z = z.real

        if self.get_figaxes().get_3d():
            crange = self._update_range(crange,
                                        (np.amin(z), np.amax(z)))
            return crange

        if (xrange[0] is not None and
            xrange[1] is not None and
            yrange[0] is not None and
                yrange[1] is not None):
            if x.size*y.size == z.size:
                idx1 = np.where((y >= yrange[0]) & (y <= yrange[1]))[0]
                idx2 = np.where((x >= xrange[0]) & (x <= xrange[1]))[0]
                if (len(idx1) == 0 or len(idx2) == 0):
                    if (crange[0] is None and
                            crange[1] is None):
                        # this is for safety maybe not necessary
                        if scale == 'log':
                            z = mask_negative(z)
                        crange = self._update_range(crange,
                                                    (np.min(z), np.max(z)))

                else:
                    zt = z[idx1, :]
                    zt = zt[:, idx2]
                    if scale == 'log':
                        zt = mask_negative(zt)
                    crange = self._update_range(crange,
                                                (np.nanmin(zt), np.nanmax(zt)))

            elif x.size == z.size and y.size == z.size:
                idx1 = np.where((y.flatten() >= yrange[0]) & (y.flatten() <= yrange[1]) &
                                (x.flatten() >= xrange[0]) & (x.flatten() <= xrange[1]))[0]
                if (len(idx1) == 0):
                    if scale == 'log':
                        zt = mask_negative(zt)
                    crange = self._update_range(crange,
                                                (min(zt), max(zt)))
                else:
                    zt = z.flatten()[idx1]
                    if scale == 'log':
                        zt = mask_negative(zt)
                    crange = self._update_range(crange,
                                                (np.amin(zt), np.amax(zt)))
            else:
                # if len(_tri) == len(z), it comes here
                crange = self._update_range(crange,
                                            (np.nanmin(z), np.nanmax(z)))

        return crange

    def set_alpha(self, value, a):
        a.set_alpha(value)
        a.set_array(a.get_array())
        self.setp('alpha', value)

    def get_alpha(self, a):
        return a.get_alpha()

    def set_interp(self, value, a):
        self.setp('interp', value)
        if value == 'cubic':
            avalue = 'bicubic'
        elif value == 'linear':
            avalue = 'bilinear'
        else:
            avalue = value

        x, y, z = self.getp(("x", "y", "z"))

        if (x.size*y.size != z.size and
                not (x.size == z.size and y.size == z.size)):
            print('FigImage: array size is wrong')
            return
        if self.get_figaxes().get_3d():
            xp, yp, zp = x, y, z
            a.set_gl_interp(value)
        else:
            xp, yp, zp = self.interp_image(x, y, z)
            a.set_array(zp)
            setattr(a.get_array(), '_xyp', (xp, yp))
        a.set_interpolation(avalue)

    def get_interp(self, a):
        return self.getp('interp')
    # needs to be implemented...

    def _update_range_gl(self, range, idx):
        for a in self._artists:
            if hasattr(a, 'get_gl_data_extent'):
                tmprange = a.get_gl_data_extent()[idx]
            range = self._update_range(range,
                                       tmprange)
        return range

    def get_zrange(self, zrange=[None, None],
                   xrange=[None, None],
                   yrange=[None, None],
                   scale='linear'):
        if self.get_figaxes().get_3d():
            zrange = self._update_range_gl(zrange, 2)
        return zrange

    def handle_axes_change(self, evt=None):
        if len(self._artists) != 0:
            a1 = self._artists[0]
            flag = True
            self.del_artist(delall=True)
            self.delp('loaded_property')
            self.generate_artist()
        else:
            flag = False
        if flag:
            a2 = self._artists[0]
            ifigure.events.SendReplaceEvent(self, a1=a1, a2=a2)

    def save_data2(self, data=None):
        def check(obj, name):
            if not isinstance(obj.getp(name), np.ndarray):
                return False
            if not isinstance(obj.getvar(name), np.ndarray):
                return False
            return obj.getp(name) is obj.getvar(name)

        if data is None:
            data = {}
        var = {'x': check(self, 'x'),
               'y': check(self, 'y'),
               'z': check(self, 'z')}

        if not var["x"]:
            if self._save_mode == 0:
                var["xdata"] = self.getp("x")
            else:
                var["xdata"] = np.array([0, 1])
        if not var["y"]:
            if self._save_mode == 0:
                var["ydata"] = self.getp("y")
            else:
                var["ydata"] = np.array([0, 1])
        if not var["z"]:
            if self._save_mode == 0:
                var["zdata"] = self.getp("z")
            else:
                var["zdata"] = np.zeros([2, 2])

#        dprint2('save_data2', var)
        data['FigImage'] = (1, var)
        data = super(FigImage, self).save_data2(data)
        return data

    def load_data2(self, data):
        d = data['FigImage']
        super(FigImage, self).load_data2(data)
        dprint2('load_data2', d[1])
        var = d[1]
        names = ["x", "y", "z"]
        for name in names:
            if var[name]:
                self.setp(name,
                          self.getvar(name))
            else:
                self.setp(name, var[name+'data'])

    def prepare_compact_savemode(self):
        var_bk = self._var.copy()
        self._var['z'] = np.zeros([2, 2])
        self._var['x'] = np.array([0, 1])
        self._var['y'] = np.array([0, 1])
        return var_bk

#    def refresh_artist_data(self):
#        print 'image refresh'
#        FigObj.refresh_artist_data(self)
#        print self._artists

    def _eval_xyz(self):
        if self.getp('use_var'):
            success = self.handle_use_var()
            if not success:
                return None, None, None
        return self.getp(("x", "y", "z"))



#
#   custom picker for ifigure
#
#   purpose of using custom picker is
#
#     to handle picking on figobj level.
#
#     to share picking routine with DnD
#     handler.
#
#     to turn on/off the picking externally
#     using the variable "active".
#
#     to call picker to page objects.
#
#
#  Author :
#         Syun'ichi Shiraiwa
#  E-mail :
#         shiraiwa@psfc.mit.edu
#
# *******************************************
#     Copyright(c) 2012- PSFC, MIT
# *******************************************
import logging
from matplotlib.lines import Line2D
from matplotlib.patches import Patch, Rectangle
from matplotlib.text import Text
from matplotlib.image import AxesImage

from ifigure.utils.cbook import Write2Main
from ifigure.ifigure_config import pick_r

import numpy as np

active = True
figchecked = False
# def AxesPicker(aritst, evt):
#    return axes_picker(artist, evt)


def label_picker(axes, evt):
    alist = [axes.title,
             axes.get_xaxis().get_label(),
             axes.get_yaxis().get_label()]
    if hasattr(axes, 'get_zaxis'):
        alist.append(axes.get_zaxis().get_label())

    for a in alist:
        hit, extra = a.contains(evt)
        if hit:
            return a
    return None


def axes_picker(artist, evt, canvas=None):
    from ifigure.mto.fig_grp import FigGrp

    if canvas is None:
        canvas = evt.guiEvent.GetEventObject()  # backend_wxagg
    tb = canvas.GetParent().toolbar  # toolbar
    ifig_canvas = canvas.GetParent()  # ifigure_canvas

    check = label_picker(artist, evt)
    if check is not None:
        return True, {"mpl_artist": check, "linked_artist": artist}
#    print 'Axex Picker', artist.figobj
    aaa = artist.get_children()
    bbb = [a for a in aaa]
    for a in bbb:
        aaa = aaa + a.get_children()
#    for a in aaa:
#       if hasattr(a, 'figobj'):  print a, a.figobj
    alist = reversed(
        sorted(
            [(a.zorder, a) for a in aaa
                if (hasattr(a, "figobj") and a in a.figobj.get_artists_for_pick())],
             key = lambda x: x[0])
    )

    # print [x[1].figobj for x in alist]
    ifig_sel = [a() for a in ifig_canvas.selection]
    for z, a in alist:
        figobj = a.figobj
        if figobj._pickmask:
            continue
        if (evt.xdata is None and
                not figobj.allow_outside()):
            continue
#           print 'checking aritst..', z, a.figobj.get_full_path()
        hit, extra = figobj.picker_a(a, evt)
        if hit:
            if isinstance(a.figobj, FigGrp):
                #                  globals()["active"]=False
                return hit, extra
            else:
                rootgrp = a.figobj.get_root_figgrp()
                if (rootgrp is not None):  # if in group
                    p = a.figobj
                    if a in ifig_sel:
                        #                          globals()["active"]=False
                        return True, {"child_artist": a}  # result in unselect
                    # go up tree till rootgrp
                    p = p.get_parent()
                    while rootgrp is p.get_root_figgrp():
                        for a in p._artists:
                            if a in ifig_sel:
                                hit, extra = p.picker_a(a, evt)
                                return hit, extra
#                              if hit:
#                                  globals()["active"]=False
#                                  return True, {"child_artist":a}
                        p = p.get_parent()
                    a = rootgrp._artists[0]
#                  globals()["active"]=False
                return True, {"child_artist": a}

    # this checks picker to axes object in amode
    # base fig_axes does have ispickable_a False,
    # which prevent it from being picked.
    # inlet axes, and its derived class can be picked.
#    if tb.ptype == 'amode':
    if True:
        if artist.figobj.ispickable_a:
            hit, extra = artist.figobj.picker_a(artist, evt)
            if hit:
                return True, {"child_artist": artist}

        # return True, {"child_artist":None}
    return False, {"child_artist": None}

# def figure_picker(figure, evt):
#    canvas = evt.guiEvent.GetEventObject()  ### backend_wxagg
#    tb = canvas.GetParent().toolbar         ### toolbar

#    if not globals()["figchecked"]:
#       hit, extra = fig_picker(figure, evt)
#       if hit:

#          globals()["active"]=False
#          globals()["figchecked"] = True
#          return True, extra
#       globals()["figchecked"] = True
#    return False, {"child_artist":None}


def figlabel_picker(fig, evt):
    if fig.figobj._title_artist is not None:
        alist = [fig.figobj._title_artist]
    else:
        alist = []
    for a in alist:
        hit, extra = a.contains(evt)
        if hit:
            return a
    return None


def fig_picker(artist, evt):
    from ifigure.mto.fig_grp import FigGrp
    from ifigure.mto.fig_axes import FigAxes

    canvas = evt.guiEvent.GetEventObject()  # backend_wxagg
    tb = canvas.GetParent().toolbar  # toolbar
    ifig_canvas = canvas.GetParent()  # ifigure_canvas

    check = figlabel_picker(artist, evt)
    if check is not None:
        return True, {"mpl_artist": check, "linked_artist": artist}

    ptype = tb.ptype
    ifig_sel = [a() for a in ifig_canvas.selection]
#    alist =reversed(
#           sorted(
#           [(a.zorder, a) for a in artist.get_children()
#            if (hasattr(a, "figobj") and not isinstance(a.figobj, FigGrp))]
#           )
#          )
    alist = reversed(
        sorted(
            [(a.zorder, a) for a in artist.get_children()
             if (hasattr(a, "figobj") and a in a.figobj.get_artists_for_pick())],
             key = lambda x: x[0])
    )
    for z, a in alist:
        flag = True
        figobj = a.figobj
        if isinstance(figobj, FigAxes):
            continue
        if hasattr(a, '_is_frameart'):
            if ifig_canvas._frameart_mode:
                if not a._is_frameart:
                    continue
            else:
                if a._is_frameart:
                    continue

        if flag:
            #           if ptype == 'pmode':
            #              hit, extra=figobj.picker(a, evt)
            #          if ptype == 'amode':
            hit, extra = figobj.picker_a(a, evt)
            if hit:
                if isinstance(a.figobj, FigGrp):
                    return hit, extra
                else:
                    rootgrp = a.figobj.get_root_figgrp()
                    if (rootgrp is not None):  # if in group
                        p = a.figobj
                        if a in ifig_sel:
                            # result in unselect
                            return True, {"child_artist": a}
                        # go up tree till rootgrp
                        p = p.get_parent()
                        while rootgrp is p.get_root_figgrp():
                            for a in p._artists:
                                if a in ifig_sel:
                                    #                          if p._artists[0] in ifig_sel:
                                    #                              a = p._artists[0]
                                    return p.picker_a(a, evt)
                            p = p.get_parent()
                        a = rootgrp._artists[0]
#                      return True, {"child_artist":a}
                # print hit to object which is not in group
                # and which is not group
                return True, {"child_artist": a}
    return False, {"child_artist": None}

# def picker(artist, evt, ptype):
    # if evt.xdata is None: return False, {}
#    if active:
#       figobj=artist.figobj

#       try:
#           if ptype == 'pmode':
#              hit, extra=figobj.picker(artist, evt)
#           if ptype == 'amode':
#              hit, extra=figobj.picker_a(artist, evt)
#       except Exception:
#            logging.exception("custom_picker: figobj.picker() failed")
#            return False, {}
#       if hit:
#           globals()["active"]=False
#       return hit, {}
#    else:
#       return False, {}


def CheckLineHit(x, y, x0, y0, trans=None, itrans=None):
    '''
    hit check between curve and pont
    if x, y are data coords.
      trans  : axes.trasnData.transform,
      itrans : axes.trasnData.inverted().transform
    '''
    def non_trans(args):
        return args

    if len(x) == 0 or len(y) == 0:
        return False, 0

    if trans is None:
        trans = non_trans
        itrans = non_trans
    if itrans is None:
        itrans = non_trans
    if not isinstance(x, np.ndarray):
        x = np.array(x)
    if not isinstance(y, np.ndarray):
        y = np.array(y)

    x0d, y0d = trans((x0, y0))

    x1dt, y1dt = itrans((x0d-pick_r, y0d-pick_r))
    x2dt, y2dt = itrans((x0d+pick_r, y0d+pick_r))

    x1d, x2d = sorted([x1dt, x2dt])
    y1d, y2d = sorted([y1dt, y2dt])

    ic = np.where((x > x1d) & (x < x2d) & (y > y1d) & (y < y2d))[0]

    if len(ic) == 0:
        ic = None
        if x.shape[0] > 255:
            return False, 0
    else:
        ic = ic[0]

    if ic is not None:
        return True, ic

    xys = np.vstack((x, y)).transpose()

    d = [np.linalg.norm(trans(xy)-[x0d, y0d]) for xy in xys]

    ic = np.argmin(d)

    if len(x) == 1:
        return False, 0

    ans = True
    if ic != len(x)-1:
        x1d, y1d = trans((x[ic], y[ic]))
        x2d, y2d = trans((x[ic+1], y[ic+1]))
        ans = ans and linehit_test(x0d, y0d, x1d, y1d, x2d, y2d)
    if ic != 0:
        x1d, y1d = trans((x[ic], y[ic]))
        x2d, y2d = trans((x[ic-1], y[ic-1]))
        ans = ans and linehit_test(x0d, y0d, x1d, y1d, x2d, y2d)
    if ic == 0:
        x1d, y1d = trans((x[0], y[0]))
        x2d, y2d = trans((x[1], y[1]))
        ans = ans and check_inner(x0d, y0d, x1d, y1d, x2d, y2d)
    if ic == len(x)-1:
        x1d, y1d = trans((x[ic], y[ic]))
        x2d, y2d = trans((x[ic-1], y[ic-1]))
        ans = ans and check_inner(x0d, y0d, x1d, y1d, x2d, y2d)

    return ans, ic


def linehit_test(x0, y0, x1, y1, x2, y2):
    #    print x0, y0, x1, y1, x2, y2
    d, l = norm_d(x0, y0, x1, y1, x2, y2)
#    print d, l
    if d > pick_r:
        return False
    return True
    #check_inner(x0, y0, x1, y1, x2, y2)


def abs_d(x0, y0, x1, y1):
    v1 = (x0-x1, y0-y1)
    l1 = np.sqrt(abs(v1[0]*v1[0]+v1[1]*v1[1]))
    return l1


def norm_d(x0, y0, x1, y1, x2, y2):
    # distance between (x0, y0) to a line
    # from (x1, y1) to (x2, y2)

    # return (normal_distance, distance between 0 and 1)

    v1 = (x0-x1, y0-y1)
    v2 = (x2-x1, y2-y1)

    l1 = np.sqrt(v1[0]*v1[0]+v1[1]*v1[1])
    l2 = np.sqrt(v2[0]*v2[0]+v2[1]*v2[1])

    if l1 == 0:
        return 0., 0.
    if l2 == 0:
        return 0., 0.

    cos = (v1[0]*v2[0]+v1[1]*v2[1])/l1/l2

    d = l1*np.sqrt(1. - cos*cos)
    return d, l1


def check_inner(x0, y0, x1, y1, x2, y2):
    #  check if (x0, y0) is inbetween (x1, y1) and (x2, y2)
    v1 = (x0-x1, y0-y1)
    v2 = (x2-x1, y2-y1)
    if (v1[0]*v2[0]+v1[1]*v2[1]) <= 0:
        return False

    v1 = (x0-x2, y0-y2)
    v2 = (x1-x2, y1-y2)
    if (v1[0]*v2[0]+v1[1]*v2[1]) <= 0:
        return False

    return True


def activate():
    globals()["active"] = True
    globals()["figchecked"] = False



from __future__ import print_function
import wx
from ifigure.utils.cbook import parseStr
from ifigure.mto.py_code import PyData
from ifigure.mto.py_code import PyCode
#from ifigure.add_on.data.mdsplus_tree import MdsplusTreeNode
import wx
import sys
import os
import re
import string
from numpy import *
from collections import OrderedDict
import ifigure.utils.mdsplusr as mds
from ifigure.utils.edit_list import DialogEditList
from netCDF4 import Dataset

######################################################
#         Setting for module file for py_module
#
#   General rule:
#      This file will be automatically loaded when
#      py_module object is created. Also, py_module
#      keeps track the file modification time. If
#      the file is updated, it will be automaticaaly
#      reloaded.


#      Strong recommendation : make module "independent".

#      Py_Modules does not check the dependency of
#      modules.
#      If moduels used in Py_Modules depends on
#      each other by for example module variable,
#      it will cause complicate  module loading
#      order-dependency problem.
#
#   name of module
module_name = 'netcdf_node'
class_name = 'netcdf_node'
#   module_evt_handler
#   functions which can be called from project tree
#
#    (menu name, function name, a flat to call skip()
#     after running function)
#
#   By default these function should return None
#   or True
#   if it return False, ifigure stops exectuion at
#   this module
#
menu = []
#
#   method lists module functions which will
#   be registered as method
#
#   spceical methods:
#      init  : called when mto will be made and
#              this module is first loaded.
#      clean : called when mto will be killed
#
method = ['isNetCDFNode', '_getitem_']

icon = 'data.png'
can_have_child = True
######################################################


def isNetCDFNode(self):
    return True


def _getitem_(self, key):
    print(key)

    def find_root(obj):
        root = obj
        while root is not None:
            try:
                if root.get_parent()._obj.isNetCDFNode():
                    root = root.get_parent()
            except Exception:
                return root.get_parent()

    obj = self.td
    root = find_root(obj)
    fpath = root.path2fullpath(pathmode='nc_pathmode',
                               path='nc_path')

    g = Dataset(fpath, 'r', format='NETCDF4')

    p = obj
    names = []
    while p != root:
        names.append(str(p.name))
        p = p.get_parent()

    for i in range(len(names)-1):
        g = getattr(g0, names[len(names)-i-1])
    g0.close()
    return g[names[0]][:]



from collections import OrderedDict
import six

# this is assigned to figplot
default_metadata_keys = ('name',
                         'long_description')

# this is assigned to figpage
default_metadata_file_keys = ('name',
                              'long_description',
                              'source',
                              'date',
                              'user_name',
                              'user_id',
                              'comment',
                              'keywords')

# this is assigned to figplot under data dictionary key.
default_metadata_data_keys = ('name',
                              'long_description',
                              'unit',
                              'range')
# In addition to this shape and type will be written
# automatically.


class MetadataHolder(object):
    def set_default_metadata(self):
        return self.setvar('metadata', OrderedDict())

    def has_metadata(self):
        return self.getvar('metadata') is not None

    def fill_default_metadatavalue(self):
        pass

    def get_metadata(self):
        return self.getvar('metadata')

    def set_metadata(self, *args):
        '''
        set_metadata({'xdata':'{'name':'',..
                      'ydata':'{'name':''...})
        set_metadata('x_axis', 'xlabel')
        '''
        if len(args) == 1:
            self.setvar('metadata', args[0])
        elif len(args) == 2:
            if self.getvar('metadata') is None:
                self.set_default_metadata()
            self.getvar('metadata')[args[0]] = args[1]

    def update_data_metadata(self, updatename=False):
        # autofill name,  range
        # if data is not give, try to get it
        metadata = self.getvar('metadata')
        for k in six.iterkeys(metadata):
            if k.startswith('data'):
                dd = metadata[k]
                for name in six.iterkeys(dd):
                    axisname = name[0]  # x, y, z, c
                    m = 'get_'+axisname+'axisparam'
                    if hasattr(self, m):
                        m = getattr(self, m)
                        param = m()
                        if updatename:
                            dd[name]['name'] = param.labelinfo[0]
                        dd[name]['range'] = str(param.range)

    def update_file_metadata(self):
        from ifigure.mto.fig_page import FigPage
        if not isinstance(self, FigPage):
            return
        from ifigure.utils.get_username import get_username
        self.getvar('metadata')['user_id'] = get_username()

    def assign_default_file_metadata(self, data=None):
        from ifigure.mto.fig_page import FigPage
        if not isinstance(self, FigPage):
            return
        if self.has_metadata():
            metadata = self.getvar('metadata')
        else:
            metadata = OrderedDict()
        for k in default_metadata_file_keys:
            if not k in metadata:
                metadata[k] = ''
        self.setvar('metadata', metadata)
        self.update_file_metadata()

    def assign_default_metadata(self, data=None):
        '''
        will be overwritten by subclass
        '''
        def fill_dataset(dd, d):
            for key in six.iterkeys(d):
                if not key in dd:
                    dd[key] = OrderedDict()
                for kk in default_metadata_data_keys:
                    if not kk in dd[key]:
                        dd[key][kk] = ''
        if self.has_metadata():
            metadata = self.getvar('metadata')
        else:
            metadata = OrderedDict()
        self.setvar('metadata', metadata)
        for key in default_metadata_keys:
            if not key in metadata:
                metadata[key] = ''
        if data is None:
            try:
                data = self.export()
            except:
                pass
        if data is not None:
            if len(data) == 1:
                if not 'data' in metadata:
                    metadata['data'] = {}
                fill_dataset(metadata['data'], data[0])
            else:
                for kk,  ddd in enumerate(data):
                    if not 'data'+str(kk+1) in metadata:
                        metadata['data'+str(kk+1)] = {}
                    fill_dataset(metadata['data'+str(kk+1)],
                                 ddd)
            self.update_data_metadata(updatename=True)



import numpy as np


def read_ishotktime(f):
    line = f.readline()
    line = line.rstrip("\r\n")

    a = line.split(' ')
    a = [x for x in a if len(x) != 0]
    return int(a[0]), int(a[1])


def read1055(f):
    '''
    form1055 = '(1x,a10,2a5)'
    '''
    line = f.readline()
    line = line.rstrip("\r\n")

    a = line[1:11]
    b = line[11:16]
    c = line[16:21]
    return a, (b, c)


def read1060(f):
    '''
    form1060=' (1x ,f7.2,10x,i5,11x,i5,1x,a3,1x,i3,1x,i3,1x,a3)'
    '''
    line = f.readline()
    line = line.rstrip("\r\n")

    a = line[1:8]
    b = line[18:23]
    c = line[34:39]
    d = line[40:43]
    e = line[44:47]
    f = line[48:51]
    g = line[52:55]
    return float(a), int(b), int(c), d, int(e), int(f), g


def read1040(f):
    '''
    form1040 = '(1x,4e16.9)'
    '''
    line = f.readline()
    line = line.rstrip("\r\n")
    line = line + ' '*45
    a = line[1:17]
    b = line[17:33]
    c = line[33:49]
    d = line[49:65]

    return a, b, c, d


def read1040f(f):
    a, b, c, d = read1040(f)
    return float(a), float(b), float(c), float(d)


def read1040_array(f, num):
    i = 0
    val = ()
    while(i < num):
        val = val + read1040(f)
        i = i+4
    return [float(x) for x in val[0:num]]


def read_afile(file=None):
    '''
       form1055 = '(1x,a10,2a5)'
       uday = ''
       mfvers = strarr(2)
       ishot = 0l
       ktime1=0l
       jflag=0l
       lflag=jflag
       limloc=0l
       mco2v=0l
       mco2r=0l
       qmflag = ''
       form1060=' (1x ,f7.2,10x,i5,11x,i5,1x,a3,1x,i3,1x,i3,1x,a3)'
       limloc = ''
       on_ioerror,abinary
       readf,2,form=form1055,uday,mfvers
       readf,2,ishot,ktime1
       on_ioerror,null
       readf,2,time
       readf,2,timejj,jflag,lflag,limloc,mco2v,mco2r,qmflag,form=form1060
       form1040 = '(1x,4e16.9)'
       readf,2,form=form1040,tsaisq,rcencm,bcentr,pasmat
       readf,2,form=form1040,cpasma,rout,zout,aout
       readf,2,form=form1040,eout,doutu,doutl,vout
       readf,2,form=form1040,rcurrt,zcurrt,qsta,betat
       readf,2,form=form1040,betap,ali,oleft,oright
       readf,2,form=form1040,otop,obott,qpsib,vertn
       rco2v = fltarr(mco2v)
       dco2v = rco2v
       rco2r = fltarr(mco2r)
       dco2r = rco2r
       readf,2,form=form1040,rco2v
       readf,2,form=form1040,dco2v
       readf,2,form=form1040,rco2r
       readf,2,form=form1040,dco2r
       readf,2,form=form1040,shearb,bpolav,s1,s2
       readf,2,form=form1040,s3,qout,olefs,orighs
       readf,2,form=form1040,otops,sibdry,areao,wplasm
       readf,2,form=form1040,terror,elongm,qqmagx,cdflux
       readf,2,form=form1040,alpha,rttt,psiref,xndnt
       rzseps = fltarr(2,2)
       readf,2,form=form1040,rzseps
       rseps = transpose(rzseps(0,*))
       zseps = transpose(rzseps(1,*))
       readf,2,form=form1040,sepexp,obots,btaxp,btaxv
       readf,2,form=form1040,aaq1,aaq2,aaq3,seplim
       readf,2,form=form1040,rmagx,zmagx,simagx,taumhd
       readf,2,form=form1040,betapd,betatd,wplasmd,fluxx
       readf,2,form=form1040,vloopt,taudia,qmerci,tavem
       ; The next bunch of things needs to know nsilop,magpri,nfcoil,nesum
       ; These guys are in parameter statements in eparmdn.for
       nfcoil = 15
       nsilop = 26
       magpri = 26
       nesum = 3
       csilop = fltarr(nsilop)
       cmpr2 = fltarr(magpri)
       readf,2,form=form1040,csilop,cmpr2
       ccbrsp = fltarr(nfcoil)
       readf,2,form=form1040,ccbrsp
       ecurrt = fltarr(nesum)
       readf,2,form=form1040,ecurrt
       readf,2,form=form1040,pbinj,rvsin,zvsin,rvsout
       readf,2,form=form1040,zvsout,vsurfa,wpdot,wbdot
       readf,2,form=form1040,slantu,slantl,zuperts,chipre
       readf,2,form=form1040,cjor95,pp95,ssep,yyy2
       readf,2,form=form1040,xmmc,cprof,oring,cjor0
       form1042 = '(1x,a42)'
       header = ''
       readf,2,form=form1042,header
       print,header 
    '''
    # file='/home/shiraiwa/a1120912012.01000'
    f = open(file, 'r')

    uday, mfvers = read1055(f)
    ishot, ktime = read_ishotktime(f)
    time = float(f.readline())
    timejj, jflag, lflag, limloc, mco2v, mco2r, qmflag = read1060(f)

    tsaisq, rcencm, bcentr, pasmat = read1040f(f)
    cpasma, rout, zout, aout = read1040f(f)
    eout, doutu, doutl, vout = read1040f(f)
    rcurrt, zcurrt, qsta, betat = read1040f(f)
    betap, ali, oleft, oright = read1040f(f)
    otop, obott, qpsib, vertn = read1040f(f)

    rco2v = read1040_array(f, mco2v)
    dco2v = read1040_array(f, mco2v)
    rco2r = read1040_array(f, mco2r)
    dco2r = read1040_array(f, mco2r)

    shearb, bpolav, s1, s2 = read1040f(f)
    s3, qout, olefs, orighs = read1040f(f)
    otops, sibdry, areao, wplasm = read1040f(f)
    terror, elongm, qqmagx, cdflux = read1040f(f)
    alpha, rttt, psiref, xndnt = read1040f(f)

    rzseps = np.array(read1040(f)).reshape((2, 2))
    rseps = np.transpose(rzseps[0, :])
    zseps = np.transpose(rzseps[1, :])

    sepexp, obots, btaxp, btaxv = read1040f(f)
    aaq1, aaq2, aaq3, seplim = read1040f(f)
    rmagx, zmagx, simagx, taumhd = read1040f(f)
    betapd, betatd, wplasmd, fluxx = read1040f(f)
    vloopt, taudia, qmerci, tavem = read1040f(f)

    val = {}

    val["uday"] = uday
    val["mfvers"] = mfvers
    val["ishot"] = ishot
    val["ktime"] = ktime
    val["time"] = time
    val["timejj"] = timejj
    val["jflag"] = jflag
    val["lflag"] = lflag
    val["limloc"] = limloc
    val["mco2v"] = mco2v
    val["mco2r"] = mco2r
    val["qmflag"] = qmflag
    val["tsaisq"] = tsaisq
    val["rcencm"] = rcencm
    val["bcentr"] = bcentr
    val["pasmat"] = pasmat
    val["cpasma"] = cpasma
    val["rout"] = rout
    val["zout"] = zout
    val["aout"] = aout
    val[" eout"] = eout
    val["doutu"] = doutu
    val["doutl"] = doutl
    val["vout"] = vout
    val["rcurrt"] = rcurrt
    val["zcurrt"] = zcurrt
    val["qsta"] = qsta
    val["betat"] = betat
    val["betap"] = betap
    val["ali"] = ali
    val["oleft"] = oleft
    val["oright"] = oright
    val["otop"] = otop
    val["obott"] = obott
    val["qpsib"] = qpsib
    val["vertn"] = vertn
    val["rco2v"] = rco2v
    val["dco2v"] = dco2v
    val["rco2r"] = rco2r
    val["dco2r"] = dco2r
    val["shearb"] = shearb
    val["bpolav"] = bpolav
    val["s1"] = s1
    val["s2"] = s2
    val["s3"] = s3
    val["qout"] = qout
    val["olefs"] = olefs
    val["orighs"] = orighs
    val["otops"] = otops
    val["sibdry"] = sibdry
    val["areao"] = areao
    val["wplasm"] = wplasm
    val["terror"] = terror
    val["elongm"] = elongm
    val["qqmagx"] = qqmagx
    val["cdflux"] = cdflux
    val["alpha"] = alpha
    val["rttt"] = rttt
    val["psiref"] = psiref
    val["xndnt"] = xndnt
    val["zseps"] = zseps
    val["rseps"] = rseps
    val["sepexp"] = sepexp
    val["obots"] = obots
    val["btaxp"] = btaxp
    val["btaxv"] = btaxv
    val["aaq1"] = aaq1
    val["aaq2"] = aaq2
    val["aaq3"] = aaq3
    val["seplim"] = seplim
    val["rmagx"] = rmagx
    val["zmagx"] = zmagx
    val["simagx"] = simagx
    val["taumhd"] = taumhd
    val["betapd"] = betapd
    val["betatd"] = betatd
    val["wplasmd"] = wplasmd
    val["fluxx"] = fluxx
    val["vloopt"] = vloopt
    val["taudia"] = taudia
    val["qmerci"] = qmerci
    val["tavem"] = tavem

    return val



from __future__ import print_function
import weakref
import traceback
import wx


class FigControl(object):
    def __init__(self):
        self._cc_callback_name = None
        self._cc_callback_obj = None

    def set_control_changed_callback(self, obj, name):
        self._cc_callback_name = name
        self._cc_callback_obj = weakref.ref(obj)

    def call_control_changed_callback(self):
        if self._cc_callback_name is None:
            return
        try:
            m = getattr(self._cc_callback_obj(), self._cc_callback_name)
        except:
            print(('callback not found', self))
            traceback.print_exc()
            return
        try:
            wx.CallAfter(m, self)
        except:
            traceback.print_exc()



''''
   this example shows how to add extra menu to mdsscope
   and add call-back routines when the menu item is 
   selected
'''

from ifigure.interactive import figure
from ifigure.mdsplus.mdsscope import MDSScope
import wx

exp = 'xtomo'
node = '.BRIGHTNESSES.ARRAY_3:CHORD_23'


class MDSScopeExtraMenuGUI(MDSScope):
    def __init__(self, *args, **kargs):
        MDSScope.__init__(self, *args, **kargs)
        extra_menu = wx.Menu()
        self.menuBar.Insert(self.menuBar.GetMenuCount()-1,
                            extra_menu, "ExtraMenu")
        self.add_menu(extra_menu, wx.ID_ANY,
                      "Plot Spectra...",
                      "plot spectra of panel #1",
                      self.onPlotSpectra)
        self.nsec(2)
        self.cls()

        self.data = self.AddSignalScript(experiment=exp,
                                         signals={'y': node},
                                         kind='plot')

    def onPlotSpectra(self, evt):
        try:
            p = self.data.get_child(0)
            x = p.getvar('x')
            y = p.getvar('y')
            self.isec(1)
            self.spec(x, y)
        except:
            import traceback
            traceback.print_exc()
        evt.Skip()


if not obj.get_parent().has_child('mdsscope_extramenu_book'):
    v = figure()
    v.book.rename('mdsscope_extramenu_book')
    v.book.set_keep_data_in_tree(True)
    v.book.Close()

obj.get_parent().mdsscope_extramenu_book.Open(MDSScopeExtraMenuGUI)



from __future__ import print_function
######################################################
#     sample "run" script for parametric solver
######################################################
#
#  this script consist from
#
#    1) modifier function
#        this function defines how to modify the
#        model based on parameter. it should accept
#        three arguments.
#           model : a model directory for each case
#           name  : a list of variables to be modified
#           value : a list of values, to which variabls
#                   are set
#
#        hint:
#          1) setting->parameters is copyed
#             to model->parameters. global variables
#             to the model should be stored here
#          2) each cases should be indpependent run.
#             this allows multi-threading of parameter
#             scan
#          3) if you are running models mutually dependent
#             prepare different solver script
#
#
#    2) main script to run parametrci scan
#        In usual usage, it is not necessary
#        to change this part.
#        this part of script is exposed to user as
#        an sample.
#
import wx
import weakref
import threading
from six.moves import queue as Queue
import ifigure.events


def modifier(worker, name, param):

    print('Running modifier')
    print(('working on model directory ...', model))
    for i in range(len(name)):
        print(('appling new parameter...', name[i], param[i]))
        worker.parameters.setvar(name[i], param[i])


renew_model_each_time = True
solver = obj.get_parent()
tmodel = solver._model
use_def_merger = solver._use_def_merger
vname = solver._pname
value = solver._pvalue
run_cases = solver._run_cases
print(run_cases)


def run_workers(workers, sol):
    def submit(worker, k, queue):
        rpath = tmodel.get_root_model().get_td_path(tmodel)
        rname = tmodel.get_root_model()._name
        rmodel = worker.get_child(name=rname)

        if (renew_model_each_time and
                rmodel is not None):
            # print "Desotry", rmodel
            rmodel.destroy()
        rmodel = worker.get_child(name=rname)
        if rmodel is None:
            tmodel.get_root_model().duplicate(worker, rname, save_script_link=True)
            rmodel = worker.get_child(name=rname)

        model = rmodel.resolve_td_path(rpath)

        solver.apply_modifier(modifier, worker, k)
        threads = model.Run(return_queue=queue)
        # print 'running ', model.get_full_path(), threads[0]
        worker._thread_name = threads[0]

    def finish_job(method, idx, w, sol, queue):
        method(idx, w, sol)
        queue.put(w)

    q = Queue.Queue()
    solver._queue = q
    finished = 0
    current = 0
    status = weakref.WeakKeyDictionary()
    work_id = weakref.WeakKeyDictionary()
    for w in workers:
        status[w] = False

    model_name = solver._model.name
    threads = {}
    aborted = False
    run_index = [i for i in range(len(run_cases)) if run_cases[i]]
    while finished < len(run_index):
        if aborted and not any(status):
            break
        for w in workers:
            if (not status[w] and
                    current < len(run_index) and not aborted):
                wx.CallAfter(submit, *(w, run_index[current], q))
#                 t = submit(solver, w, current, q)[0]
                work_id[w] = run_index[current]
                status[w] = True
#                 else:
#                    finished = finished + 1
                current = current + 1

        try:
            t = q.get(True)
            # t = thread name  : thread done
            #     worker       : finish job done
            #     'abort job'  : abort parametric scan
            for o in workers:
                if o._thread_name == t:
                    w = o
                    break
        except Queue.Empty:
            print(('not respond', finished))
            continue
        if t == 'abort job':
            abroted = True
        if t in workers:
            status[t] = False
            finished = finished + 1
#             print finished, sum(run_index)):
            continue

        # print 'response from', w, threading.current_thread()
        rpath = tmodel.get_root_model().get_td_path(tmodel)
        rname = tmodel.get_root_model()._name
        rmodel = w.get_child(name=rname)
        model = rmodel.resolve_td_path(rpath)

        if model._finish_script == '':
            model._finish_script = solver.merge_sol.get_full_path()
        # print 'calling merge_sol', model.do_finish
        wx.CallAfter(finish_job, model.do_finish, work_id[w], w, sol, q)
#        status[w] = False
#        finished = finished + 1

    solver._queue = None


task = obj.getvar("task")

print('### starting parametric solver ###')
solver.print_cases()
if task < 1:
    stop()

print('### apply modifier             ###')

if task < 2:
    stop()

print('### run all models             ###')
sol_base = solver._sol
rpath = tmodel.get_root_model().get_td_path(tmodel)
rname = tmodel.get_root_model()._name

if solver._num_worker < 1:
    for k, xxx in enumerate(sol_base.get_children()):
        name, case_xxx = xxx
        rmodel = case_xxx.get_child(name=rname)
        model = rmodel.resolve_td_path(rpath)
        print((case_xxx, model))
        if model is not None:
            solver.apply_modifier(modifier, case_xxx, k)
            model.Run()
else:
    workers = solver.get_workers()
    if len(workers) == 0:
        print('no workers')
        stop()
    if solver._queue is not None:
        print('there is a waiting queue, exiting...')
        stop()
    for w in workers:
        w._thread_name = ''
    t = threading.Thread(target=run_workers, args=(workers, sol_base))
    wx.CallAfter(t.start)



'''
   find_contours based on scikit-image (marching square alogrith)
'''
import numpy as np
try:
   from skimage import measure
except ImportError:
   assert False, "Can not import skimage. Consider pip install scikit-image."

def findthem(data, level, xgrid=None, ygrid=None, transpose=True):
    if transpose:
        data = np.transpose(data)
    
    segs = measure.find_contours(data, level)

    ret = []

    if xgrid is not None:
        xx = np.arange(len(xgrid))
    if ygrid is not None:
        yy = np.arange(len(ygrid))

    for s in segs:
        if xgrid is not None:
            s[:, 0] = np.interp(s[:, 0], xx, xgrid)
        if ygrid is not None:
            s[:, 1] = np.interp(s[:, 1], yy, ygrid)
        ret.append(s)

    return ret

class Cntr():
    '''
    this class mimic Cntr in LegacyContour which seems stop working
    with new version of Matplotlib
    '''
    def __init__(self, x, y, z):
        x = np.array(x, copy=False)
        if len(x.shape) == 2:
            x = x[0, :]
        y = np.array(y, copy=False)
        if len(y.shape) == 2:
            y = y[:, 0]
            
        self.x = x
        self.y = y
        self.z = np.transpose(z)
        
    def trace(self, level):
        ret = findthem(self.z,
                       level,
                       xgrid=self.x,
                       ygrid=self.y,
                       transpose=False)

        ret = ret + [None]*len(ret)
        return ret
    
def contour(*args, **kwargs):
    '''
    countour plot using findthem (above).
    mostly meant for testing
    '''
    if len(args) == 3:
        x, y, z = args
    elif len(args) == 1:
        z = args[0]
        x = None
        y = None

    data = np.transpose(z)
    nlevels = kwargs.pop('nlevels', 15)

    from ifigure.interactive import figure

    levels = np.linspace(np.min(z), np.max(z), nlevels)

    v = figure()
    v.update(False)
    for l in levels:
        segs = findthem(data, l, xgrid=x, ygrid=y, transpose=False)
        for s in segs:
            v.plot(s[:,0], s[:,1])
            
    v.update(True)



'''

  filed_ndarray 
  
  to transfer a large  data between processes.
  instead of pushing the data into pipe, 
  it creates files with two resolutions.
  main process will open coarse resolution file
  first.

  FileNDArray : used in sub process
  FileNDArrayAutoDel : used in main process, allowing
                       automatic delete of file


'''
import traceback
import tempfile
import numpy
import numpy
import os
import time
import ifigure.utils.debug as debug
debug.debug_default_level = 1
dprint1, dprint2, dprint3 = debug.init_dprints('FigNDArray')


def random_tmp_name(seed='1'):
    # generate a random string for an initial work dir
    from datetime import datetime
    import hashlib
    strtime = datetime.now().strftime("%Y-%m-%d %H:%M:%S. %f")
    # print strtime
    m = hashlib.md5()
    m.update(strtime+str(seed))
    txt = m.hexdigest()
    return txt


class FiledNDArray(object):
    def __init__(self, array):
        self._decimation = False
        dprint3('Called here')
#        fid = tempfile.NamedTemporaryFile(delete=False)
#        name = fid.name
#        name = '/home/shiraiwa/my_tmp/'+random_tmp_name()
#        fid = open(name, 'w')
        fid, name = tempfile.mkstemp()
        self.dtype = str(array.dtype)
        self._level = 0
        self._level_size = {0: array.size}
        self._level_name = {0: name}
#        print 'Storing to', name
        array.tofile(name)
        os.close(fid)  # to close tempfile here
#        fid.close()
        if array.size > 2e4 and len(array.shape) == 1:
            #           fid = tempfile.NamedTemporaryFile(delete=False)
            #           name = fid.name
            #           name = '/home/shiraiwa/my_tmp/'+random_tmp_name()
            #           fid = open(name, 'w')
            fid, name = tempfile.mkstemp()
            array2 = self._decimate(array)
            dprint3('Storing to', name)
            array2.tofile(name)
            os.close(fid)  # to close tempfile here
#           fid.close()
            self._level = 1
            self._level_size[1] = array2.size
            self._level_name[1] = name

    def restore(self, level=None):
        #        return range(1000)
        if level is None:
            level = self._level
        try:
            dprint2('Restoring' +
                    self._level_name[level].__repr__())
            val = numpy.fromfile(self._level_name[level],
                                 dtype=self.dtype)
            os.remove(self._level_name[level])
            del self._level_name[level]
        except:
            dprint2('Failed to restore the file' +
                    self._level_name[level].__repr__())
            raise
        return val

    def _filename(self, level=None):
        if level is None:
            level = 0
        return self.name+'_'+str(level)

    def _decimate(self, x):
        #        print x.shape
        chunksize = 4000
        numchunks = x.size//chunksize
        if x.size//numchunks > chunksize:
            chunksize = x.size//numchunks
#        if numchunks < 100:
#        numchunks = 2000
#        chunksize = x.size/numchunks

        xchunks = x[:chunksize*numchunks].reshape((-1, numchunks))
#        print numpy.nanmax(xchunks, axis=1)
#        print numpy.nanmax(xchunks, axis=1).shape
#        print numchunks, chunksize
        x3 = numpy.vstack((numpy.nanmin(xchunks, axis=1), numpy.nanmax(
            xchunks, axis=1))).transpose().reshape(chunksize*2)
#        print x3.shape
        # append left over..
        if chunksize*numchunks < x.size:
            x2 = numpy.hstack((x3, numpy.nanmin(
                x[chunksize*numchunks:]), numpy.nanmax(x[chunksize*numchunks:])))
            return x2
        else:
            return x3

##
# add auto delete to filed_ndarray
##
# used in main process, so that it will delete the fiel
# when it is done.


class FiledNDArrayAutoDel(FiledNDArray):
    def __del__(self):
        for level in self._level_name:
            if os.path.exists(self._level_name[level]):
                os.remove(self._level_name[level])



from __future__ import print_function
#
#   MDS worker pool for threaded worker and multiprocessing worker.
#
from ifigure.mdsplus.mds_mpth_directworker import MDSMPDirectWorker, MDSTHDirectWorker
from ifigure.mdsplus.mds_mpth_proxyworker import MDSMPProxyWorker, MDSTHProxyWorker
import multiprocessing as mp
import time
import weakref
import threading
from six.moves import queue as Queue
import ifigure.utils.debug as debug
debug.debug_default_level = 1
dprint1, dprint2, dprint3 = debug.init_dprints('MDSMPWorkerPool')
jllock = threading.Lock()


class MDSWorkerPoolBase(object):
    job_id = 0
    job_list = {}

    def submit_ana_group(self, ana_group):
        job_id = MDSWorkerPoolBase.job_id

#       if hasattr(ana_group[0],'fig_mds'):
#           print 'submit', ana_group[0].fig_mds()
        jllock.acquire()
        MDSWorkerPoolBase.job_list[job_id] = [len(ana_group),
                                              weakref.ref(ana_group)]
        self.tasks.put((job_id, [(a.jobs, a.jobnames)
                                 for a in ana_group], ana_group.global_tdi))
        jllock.release()
        MDSWorkerPoolBase.job_id = job_id + 1

    def check_ana_group_done(self):
        from ifigure.mdsplus.mdsscope import GlobalCounter

        try:
            #           print 'waiting result'
            v = self.results.get(True)
        except:
            import traceback
            print(traceback.format_exc())
            return None
        self.results.task_done()
        if not (isinstance(v, tuple) and len(v) == 3):
            return v
        job_id, idx, result = v
        jllock.acquire()
        if not job_id in MDSWorkerPoolBase.job_list:
            print(' should not come here,, but just in case it ')
            return None
        jllock.release()
        if job_id in MDSWorkerPoolBase.job_list:
            xxx = MDSWorkerPoolBase.job_list[job_id]
            try:
                #                for a, v in zip(ana_group, result):
                #                    a.set_result(v != None, v)
                ana_group = xxx[1]()
                if ana_group is None:
                    return None
                ana_group[idx].set_result(result != None, result)
            except:
                import traceback
                print(traceback.format_exc())
                return None
            xxx[0] = xxx[0]-1
            if xxx[0] == 0:
                del MDSWorkerPoolBase.job_list[job_id]
            return ana_group, idx
        #   should not come here,, but just in case it
        return None

    def abort_job(self):
        tasks = []
        while not self.tasks.empty():
            try:
                j = self.tasks.get(False)
                self.tasks.task_done()
                tasks.append(j)
            except:
                import traceback
                traceback.print_exc()
                break
        ana_groups = []
        for job_id, job, res in tasks:
            if job_id in MDSWorkerPoolBase.job_list:
                xxx = MDSWorkerPoolBase.job_list[job_id]
                ana_group = xxx[1]()
                ana_groups.append(ana_group)
                del MDSWorkerPoolBase.job_list[job_id]
        print(('cancelled jobs', ana_groups))
        return ana_groups

    def has_noresult(self):
        return self.results.empty()

    def is_alive(self, w_id):
        return self.workers[w_id].is_alive()

    def is_any_alive(self):
        return any([w.is_alive() for w in self.workers])

    def terminate_all(self):
        print('terminating all')
        num_alive = 0
        for w in self.workers:
            if w.is_alive():
                num_alive = num_alive + 1
        for x in range(num_alive):
            self.tasks.put(-1)
        self.tasks.join()
        print('joined')
        # stops the listening thread
        self.results.put('stop')

    def get_num_worker(self):
        return len(self.workers)


class MDSMPWorkerPool(MDSWorkerPoolBase):
    def __init__(self, num, *args, **kwargs):
        self.num = num
        MDSWorkerPoolBase.job_id = 0
        MDSWorkerPoolBase.job_list = {}

        type = args[0]
        if len(args) > 1:
            host = args[1]
            port = args[2]
        else:
            host = None
            port = None

        self.tasks = mp.JoinableQueue()
#       self.results= mp.Queue()
        self.results = mp.JoinableQueue()

        self.workers = [None]*num
        for i in range(num):
            if type == 'direct':
                w = MDSMPDirectWorker(
                    self.tasks, self.results, kwargs['translater'])
            else:
                w = MDSMPProxyWorker(
                    self.tasks, self.results, str(host), str(port))
            self.workers[i] = w
            time.sleep(0.1)
        for w in self.workers:
            w.start()


class MDSTHWorkerPool(MDSWorkerPoolBase):
    def __init__(self, num, *args, **kwargs):
        self.num = num
        MDSWorkerPoolBase.job_id = 0
        MDSWorkerPoolBase.job_list = {}

        type = args[0]
        if len(args) > 1:
            host = args[1]
            port = args[2]
        else:
            host = None
            port = None

        self.tasks = Queue.Queue()
        self.results = Queue.Queue()

        self.workers = [None]*num
        for i in range(num):
            if type == 'direct':
                w = MDSTHDirectWorker(
                    self.tasks, self.results, kwargs['translater'])
            else:
                w = MDSTHProxyWorker(
                    self.tasks, self.results, str(host), str(port))
            self.workers[i] = w
            time.sleep(0.1)
        for w in self.workers:
            w.start()



from __future__ import print_function

'''
fake listner thread for debugging
'''

import multiprocessing as mp
import threading
from six.moves import queue as Queue
import traceback
import time
from weakref import WeakKeyDictionary
import __future__

import MDSplus
from MDSplus import Connection
use_event_listener = True


class EventListenerProc(mp.Process):
    def __init__(self, task_queue, result_queue, *args, **kargs):
        super(EventListenerProc, self).__init__(*args, **kargs)
        self.task_queue = task_queue
        self.result_queue = result_queue

    def run(self, *args, **kargs):
        while True:
            #            while True:
            try:
                jobcode, jobparam = self.task_queue.get(True)  # True, 0.1)
            except:
                continue
                pass
            # jobset
            ###   (jobcode, jobparam)
            # jobcode : 'wait' : launch thread if it is not waiting
            # jobcode : 'exit' : exit loop
            # jobcode : 'send' :  when event is recieved
            if jobcode == 'wait':
                pass
            elif jobcode == 'exit':
                #                print('exiting')
                break
            elif jobcode == 'disconnect':
                pass
            else:
                pass


class EventListener(threading.Thread):
    def __init__(self, *args, **kargs):
        super(EventListener, self).__init__(*args, **kargs)
        self.queue = mp.Queue()  # queue to receive message
        self.task_queue = mp.JoinableQueue()
        self.listener_proc = EventListenerProc(self.task_queue,
                                               self.queue)
        self.listener_proc.start()

    def run(self, *args, **kargs):
        while True:
            event_name = self.queue.get(True)
            if event_name == 'stop':
                #globals()['listener_thread'] = None
                return
            mds_event_listener_lock.acquire()
            check = False
            # do nothing
            mds_event_listener_lock.release()


if __name__ == '__main__':
    server = 'ssh://shiraiwa@alcdata.psfc.mit.edu'  # this makes a dead-lokc
#    server = 'alcdata.psfc.mit.edu'                # this works
    c = Connection(server)

    listener_thread = EventListener()
    listener_thread.task_queue.put(('exit', ''), False)
    listener_thread.queue.put('stop')

    print('good bye')
